# This file consolidates the expectations of what properties/QCVariables
#  a method should produce and what in practice a QC program does produce.

__all__ = [
    "contractual_current",
    "contractual_hf",
    "contractual_mp2",
    "contractual_mp2p5",
    "contractual_mp3",
    "contractual_mp4_prsdq_pr",
    "contractual_mp4",
    "contractual_zapt2",
    "contractual_cisd",
    "contractual_qcisd",
    "contractual_qcisd_prt_pr",
    "contractual_fci",
    "contractual_remp2",
    "contractual_lccd",
    "contractual_lccsd",
    "contractual_cepa_pr1_pr",
    "contractual_cepa_pr3_pr",
    "contractual_acpf",
    "contractual_aqcc",
    "contractual_ccd",
    "contractual_bccd",
    "contractual_cc2",
    "contractual_ccsd",
    "contractual_ccsdpt_prccsd_pr",
    "contractual_ccsd_prt_pr",
    "contractual_accsd_prt_pr",
    "contractual_bccd_prt_pr",
    "contractual_cc3",
    "contractual_ccsdt",
    "contractual_ccsdt1a",
    "contractual_ccsdt1b",
    "contractual_ccsdt2",
    "contractual_ccsdt3",
    "contractual_ccsdt_prq_pr",
    "contractual_ccsdtq",
    "contractual_omp2",
    "contractual_omp2p5",
    "contractual_omp3",
    "contractual_oremp2",
    "contractual_olccd",
    "contractual_occd",
    "contractual_occd_prt_pr",
    "contractual_aoccd_prt_pr",
    "contractual_dft_current",
    "contractual_dhdft_current",
    "query_qcvar",
    "query_has_qcvar",
]


from typing import Any, Tuple

from qcengine.programs.qcvar_identities_resources import qcvars_to_atomicproperties


def query_qcvar(obj: Any, pv: str) -> Any:
    """Uniform interface to value of variable `pv` or its QCSchema alias in QCVariable store `obj`."""

    try:
        # psi4.core (module -- P::e.globals)
        # psi4.Wavefunction
        # qcdb (module)
        vval = obj.variable(pv)
    except AttributeError:
        try:
            # qcdb jobrec["qcvars"] qcel.Datum
            vval = obj[pv].data
        except (AttributeError, KeyError):
            # qcel.AtomicResult.extras["qcvars"]
            vval = obj.get(pv)
            if vval is None:
                # qcel.AtomicResult.properties dict
                vval = obj.get(qcvars_to_atomicproperties[pv])
        except TypeError:
            # qcel.AtomicResult.properties object
            vval = getattr(obj, qcvars_to_atomicproperties[pv])

    return vval


def query_has_qcvar(obj: Any, pv: str) -> bool:
    """Uniform interface to whether variable `pv` or its QCSchema alias in QCVariable store `obj`."""

    try:
        bval = obj.has_variable(pv)
    except AttributeError:
        bval = pv in obj

        if not bval:
            try:
                bval = qcvars_to_atomicproperties[pv] in obj
            except KeyError:
                bval = False

    return bval


_contractual_docstring = """
    Parameters
    ----------
    qc_module
        The program or subprogram running the job (e.g., "cfour" or "cfour-ecc").
    driver
        {"energy", "gradient", "hessian"}
        The derivative level that should be expected.
    reference
        {"rhf", "uhf", "rohf"}
        The SCF reference since programs often output differently based on it.
    method
        The target AtomicInput.model.method since "free" methods may not always be
        output (e.g., MP2 available when target is MP2 but not when target is CCSD).
    corl_type
        {"conv", "df", "cd"}
        The algorithm for the target method since programs often output differently
        based on it.
    fcae
        {"ae", "fc"}
        The all-electron vs. frozen-orbital aspect.

    Returns
    -------
    (rpv, pv, expected)
        Of all the QCVariables `pv` that should be available, returns tuple of
        whether `expected` and what key `rpv` in the reference `pv` should match.

"""


def contractual_current(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    """Given the target method, returns the CURRENT QCVariables that should be produced.

    {_contractual_docstring}
    """
    contractual_qcvars = [
        ("HF TOTAL ENERGY", "SCF TOTAL ENERGY"),
        ("HF TOTAL ENERGY", "CURRENT REFERENCE ENERGY"),
        (f"{method.upper()} CORRELATION ENERGY", "CURRENT CORRELATION ENERGY"),
        (f"{method.upper()} TOTAL ENERGY", "CURRENT ENERGY"),
    ]
    if driver == "gradient":
        contractual_qcvars.append((f"{method.upper()} TOTAL GRADIENT", "CURRENT GRADIENT"))
    elif driver == "hessian":
        # contractual_qcvars.append((f"{method.upper()} TOTAL GRADIENT", "CURRENT GRADIENT"))
        contractual_qcvars.append((f"{method.upper()} TOTAL HESSIAN", "CURRENT HESSIAN"))

    for rpv, pv in contractual_qcvars:
        expected = True
        if method == "hf" and rpv == f"{method.upper()} CORRELATION ENERGY":
            expected = False

        yield (rpv, pv, expected)


def contractual_hf(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    """Of the list of QCVariables an ideal HF should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """

    contractual_qcvars = [
        ("HF TOTAL ENERGY", "HF TOTAL ENERGY"),
        ("HF TOTAL ENERGY", "SCF TOTAL ENERGY"),
    ]
    if driver == "gradient" and method == "hf":
        contractual_qcvars.append(("HF TOTAL GRADIENT", "HF TOTAL GRADIENT"))
        # contractual_qcvars.append(("HF TOTAL GRADIENT", "SCF TOTAL GRADIENT"))
    elif driver == "hessian" and method == "hf":
        # contractual_qcvars.append(("HF TOTAL GRADIENT", "HF TOTAL GRADIENT"))
        contractual_qcvars.append(("HF TOTAL HESSIAN", "HF TOTAL HESSIAN"))
        # contractual_qcvars.append(("HF TOTAL GRADIENT", "SCF TOTAL GRADIENT"))

    for rpv, pv in contractual_qcvars:
        yield (rpv, pv, True)


def contractual_mp2(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    """Of the list of QCVariables an ideal MP2 should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "MP2 CORRELATION ENERGY",
        "MP2 TOTAL ENERGY",
        "MP2 SAME-SPIN CORRELATION ENERGY",
        "MP2 SINGLES ENERGY",
        "MP2 DOUBLES ENERGY",
        "MP2 OPPOSITE-SPIN CORRELATION ENERGY",
    ]
    if driver == "gradient" and method == "mp2":
        contractual_qcvars.append("MP2 TOTAL GRADIENT")
    elif driver == "hessian" and method == "mp2":
        # contractual_qcvars.append("MP2 TOTAL GRADIENT")
        contractual_qcvars.append("MP2 TOTAL HESSIAN")

    for pv in contractual_qcvars:
        expected = True
        if (
            (
                (
                    (qc_module == "cfour" and reference == "rohf" and method == "mp2" and driver == "hessian")
                    or (
                        qc_module in ["gamess-serial", "gamess-ddi"]
                        and reference in ["uhf", "rohf"]
                        and method == "mp2"
                    )
                    or (
                        qc_module in ["gamess-serial", "gamess-ims"]
                        and reference == "rhf"
                        and method == "mp2"
                        and driver in ["gradient", "hessian"]
                    )
                    or (
                        qc_module == "gamess"
                        and reference in ["rhf"]
                        and method in ["lccd", "ccd", "ccsd", "ccsd+t(ccsd)", "ccsd(t)"]
                    )
                    or (qc_module == "nwchem-tce" and method in ["mp2", "mp3", "mp4"])
                    or (
                        qc_module == "nwchem-cc"
                        and reference in ["rhf"]
                        and method in ["ccsd", "ccsd+t(ccsd)", "ccsd(t)"]
                    )
                    or (qc_module == "nwchem-directmp2" and reference == "rhf" and method == "mp2")
                    or (
                        qc_module == "psi4-occ"
                        and reference == "rhf"
                        and corl_type in ["df", "cd"]
                        and method
                        in [
                            "mp2",
                            "mp2.5",
                            "mp3",
                            "remp2",
                            "lccd",
                            "ccd",
                            "ccsd",
                            "ccsd(t)",
                            "a-ccsd(t)",
                            "omp2",
                            "omp2.5",
                            "omp3",
                            "oremp2",
                            "olccd",
                            "occd",
                            "occd(t)",
                            "a-occd(t)",
                        ]
                    )
                    or (
                        qc_module == "psi4-mrcc"
                        and reference in ["rhf", "uhf"]
                        and method
                        in ["ccsd", "ccsd(t)", "a-ccsd(t)", "cc3", "ccsdt-1a", "ccsdt-1b", "ccsdt-3", "ccsdt"]
                    )
                )
                and pv in ["MP2 SAME-SPIN CORRELATION ENERGY", "MP2 OPPOSITE-SPIN CORRELATION ENERGY"]
            )
            or (
                (
                    (qc_module == "psi4-detci" and method in ["mp2", "mp3", "mp4"])
                    or (
                        qc_module == "qchem" and method == "mp2"
                    )  # for structured -- can probably get these from parsing
                )
                and pv
                in [
                    "MP2 SAME-SPIN CORRELATION ENERGY",
                    "MP2 OPPOSITE-SPIN CORRELATION ENERGY",
                    "MP2 SINGLES ENERGY",
                    "MP2 DOUBLES ENERGY",
                ]
            )
            or (
                (
                    (
                        qc_module == "psi4-occ"
                        and reference == "rohf"
                        and corl_type == "conv"
                        and method in ["omp2", "omp2.5", "omp3", "oremp2", "olccd"]
                    )
                )
                and pv
                in [
                    "MP2 CORRELATION ENERGY",
                    "MP2 TOTAL ENERGY",
                    "MP2 SINGLES ENERGY",
                ]
            )
            or (
                (
                    (qc_module == "psi4-ccenergy" and reference == "rohf" and method == "ccsd" and sdsc == "sd")
                    # next two are an evasion as possibly collectable
                    or (qc_module == "cfour-vcc" and reference in ["rhf", "uhf"] and method == "bccd")
                    or (qc_module == "cfour-vcc" and reference in ["rhf"] and method == "bccd(t)")
                    or (
                        qc_module == "nwchem-tce"
                        and method
                        in ["qcisd", "lccd", "lccsd", "ccd", "cc2", "ccsd", "ccsd+t(ccsd)", "ccsd(t)", "ccsdt"]
                    )
                    or (qc_module == "gamess" and reference == "rohf" and method == "ccsd")
                    or (
                        qc_module.startswith("cfour")
                        and reference == "rohf"
                        and method in ["lccsd", "ccd", "ccsd", "ccsd(t)", "ccsdt"]
                        and sdsc == "sd"
                    )  # this is a cop out as c4 perfectly able to produce good rohf mp2 but not with same orbitals as ref definition on ccsd
                    or (
                        qc_module == "psi4-mrcc"
                        and reference in ["rohf"]
                        and method in ["ccsd", "ccsd(t)", "a-ccsd(t)", "ccsdt"]
                    )
                )
                and pv
                in [
                    "MP2 CORRELATION ENERGY",
                    "MP2 TOTAL ENERGY",
                    "MP2 SAME-SPIN CORRELATION ENERGY",
                    "MP2 OPPOSITE-SPIN CORRELATION ENERGY",
                    "MP2 SINGLES ENERGY",
                    "MP2 DOUBLES ENERGY",
                ]
            )
        ):
            expected = False

        yield (pv, pv, expected)


#        # TODO check CUSTOM SCS-MP2 _absent_


def contractual_mp2p5(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    f"""Of the list of QCVariables an ideal MP2.5 should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "MP2.5 CORRELATION ENERGY",
        "MP2.5 TOTAL ENERGY",
        "MP2.5 SAME-SPIN CORRELATION ENERGY",
        "MP2.5 SINGLES ENERGY",
        "MP2.5 DOUBLES ENERGY",
        "MP2.5 OPPOSITE-SPIN CORRELATION ENERGY",
    ]
    if driver == "gradient" and method == "mp2.5":
        contractual_qcvars.append("MP2.5 TOTAL GRADIENT")
    elif driver == "hessian" and method == "mp2.5":
        # contractual_qcvars.append("MP2.5 TOTAL GRADIENT")
        contractual_qcvars.append("MP2.5 TOTAL HESSIAN")

    for pv in contractual_qcvars:
        expected = True
        if (
            (
                (qc_module.startswith("cfour") and method in ["mp3", "mp4(sdq)", "mp4"])
                or (
                    qc_module == "psi4-occ"
                    and reference == "rhf"
                    and corl_type in ["df", "cd"]
                    and method in ["mp2.5", "mp3", "omp2.5"]
                )
                or (qc_module.startswith("nwchem") and method in ["mp3", "mp4"])
            )
            and pv in ["MP2.5 SAME-SPIN CORRELATION ENERGY", "MP2.5 OPPOSITE-SPIN CORRELATION ENERGY"]
        ) or (
            (
                (qc_module == "psi4-detci" and method in ["mp3", "mp4"])
                or (
                    qc_module == "psi4-occ"
                    and reference == "rohf"
                    and corl_type in ["conv", "df", "cd"]
                    and method in ["omp2.5", "omp3"]
                )
            )
            # Note SS/OS might be obtainable but no reference to verify
            and pv
            in [
                "MP2.5 CORRELATION ENERGY",
                "MP2.5 TOTAL ENERGY",
                "MP2.5 SAME-SPIN CORRELATION ENERGY",
                "MP2.5 OPPOSITE-SPIN CORRELATION ENERGY",
                "MP2.5 SINGLES ENERGY",
                "MP2.5 DOUBLES ENERGY",
            ]
        ):
            expected = False

        yield (pv, pv, expected)


def contractual_mp3(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    """Of the list of QCVariables an ideal MP3 should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "MP3 CORRELATION ENERGY",
        "MP3 TOTAL ENERGY",
        "MP3 SAME-SPIN CORRELATION ENERGY",
        "MP3 SINGLES ENERGY",
        "MP3 DOUBLES ENERGY",
        "MP3 OPPOSITE-SPIN CORRELATION ENERGY",
    ]
    if driver == "gradient" and method == "mp3":
        contractual_qcvars.append("MP3 TOTAL GRADIENT")
    elif driver == "hessian" and method == "mp3":
        # contractual_qcvars.append("MP3 TOTAL GRADIENT")
        contractual_qcvars.append("MP3 TOTAL HESSIAN")

    for pv in contractual_qcvars:
        expected = True
        if (
            (
                (
                    (qc_module.startswith("cfour") and method in ["mp3", "mp4(sdq)", "mp4"])
                    or (
                        qc_module == "psi4-occ"
                        and reference == "rhf"
                        and corl_type in ["df", "cd"]
                        and method in ["mp2.5", "mp3", "omp2.5", "omp3"]
                    )
                    or (qc_module.startswith("nwchem") and method in ["mp3", "mp4"])
                )
                and pv in ["MP3 SAME-SPIN CORRELATION ENERGY", "MP3 OPPOSITE-SPIN CORRELATION ENERGY"]
            )
            or (
                ((qc_module == "psi4-detci" and method in ["mp3", "mp4"]))
                and pv
                in [
                    "MP3 SAME-SPIN CORRELATION ENERGY",
                    "MP3 OPPOSITE-SPIN CORRELATION ENERGY",
                    "MP3 SINGLES ENERGY",
                    "MP3 DOUBLES ENERGY",
                ]
            )
            or (
                (
                    (
                        qc_module == "psi4-occ"
                        and reference == "rohf"
                        and corl_type in ["conv", "df", "cd"]
                        and method in ["omp2.5", "omp3"]
                    )
                )
                # Note SS/OS might be obtainable but no reference to verify
                and pv
                in [
                    "MP3 CORRELATION ENERGY",
                    "MP3 TOTAL ENERGY",
                    "MP3 SAME-SPIN CORRELATION ENERGY",
                    "MP3 SINGLES ENERGY",
                    "MP3 DOUBLES ENERGY",
                    "MP3 OPPOSITE-SPIN CORRELATION ENERGY",
                ]
            )
        ):
            expected = False

        yield (pv, pv, expected)


def contractual_mp4_prsdq_pr(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    """Of the list of QCVariables an ideal MP4(SDQ) should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "MP4(SDQ) CORRELATION ENERGY",
        "MP4(SDQ) TOTAL ENERGY",
    ]
    if driver == "gradient" and method == "mp4(sdq)":
        contractual_qcvars.append("MP4(SDQ) TOTAL GRADIENT")
    elif driver == "hessian" and method == "mp4(sdq)":
        # contractual_qcvars.append("MP4(SDQ) TOTAL GRADIENT")
        contractual_qcvars.append("MP4(SDQ) TOTAL HESSIAN")

    for pv in contractual_qcvars:
        expected = True
        if (
            (qc_module.startswith("psi4-detci") and method == "mp4")
            or (qc_module.startswith("nwchem") and method == "mp4")
        ) and pv in ["MP4(SDQ) TOTAL ENERGY", "MP4(SDQ) CORRELATION ENERGY"]:
            expected = False

        yield (pv, pv, expected)


def contractual_mp4(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    """Of the list of QCVariables an ideal MP4 should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "MP4(T) CORRECTION ENERGY",
        "MP4 CORRELATION ENERGY",
        "MP4 CORRECTION ENERGY",
        "MP4 TOTAL ENERGY",
    ]
    if driver == "gradient" and method == "mp4":
        contractual_qcvars.append("MP4 TOTAL GRADIENT")
    elif driver == "hessian" and method == "mp4":
        # contractual_qcvars.append("MP4 TOTAL GRADIENT")
        contractual_qcvars.append("MP4 TOTAL HESSIAN")

    for pv in contractual_qcvars:
        expected = True
        if (
            (qc_module.startswith("psi4-detci") and method == "mp4")
            or (qc_module.startswith("nwchem") and method == "mp4")
        ) and pv in ["MP4(T) CORRECTION ENERGY"]:
            expected = False

        yield (pv, pv, expected)


def contractual_zapt2(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    f"""Of the list of QCVariables an ideal ZAPT2 should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "ZAPT2 CORRELATION ENERGY",
        "ZAPT2 TOTAL ENERGY",
    ]
    if driver == "gradient" and method == "zapt2":
        contractual_qcvars.append("ZAPT2 TOTAL GRADIENT")
    elif driver == "hessian" and method == "zapt2":
        # contractual_qcvars.append("ZAPT2 TOTAL GRADIENT")
        contractual_qcvars.append("ZAPT2 TOTAL HESSIAN")

    for pv in contractual_qcvars:
        expected = True

        yield (pv, pv, expected)


def contractual_cisd(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    f"""Of the list of QCVariables an ideal CISD should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "CISD CORRELATION ENERGY",
        "CISD TOTAL ENERGY",
    ]
    if driver == "gradient" and method == "cisd":
        contractual_qcvars.append("CISD TOTAL GRADIENT")
    elif driver == "hessian" and method == "cisd":
        # contractual_qcvars.append("CISD TOTAL GRADIENT")
        contractual_qcvars.append("CISD TOTAL HESSIAN")

    for pv in contractual_qcvars:
        expected = True

        yield (pv, pv, expected)


def contractual_qcisd(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    f"""Of the list of QCVariables an ideal QCISD should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "QCISD CORRELATION ENERGY",
        "QCISD TOTAL ENERGY",
    ]
    if driver == "gradient" and method == "qcisd":
        contractual_qcvars.append("QCISD TOTAL GRADIENT")
    elif driver == "hessian" and method == "qcisd":
        # contractual_qcvars.append("QCISD TOTAL GRADIENT")
        contractual_qcvars.append("QCISD TOTAL HESSIAN")

    for pv in contractual_qcvars:
        expected = True

        yield (pv, pv, expected)


def contractual_qcisd_prt_pr(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    f"""Of the list of QCVariables an ideal QCISD(T) should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "QCISD(T) CORRECTION ENERGY",
        "QCISD(T) CORRELATION ENERGY",
        "QCISD(T) TOTAL ENERGY",
    ]
    if driver == "gradient" and method == "qcisd(t)":
        contractual_qcvars.append("QCISD(T) TOTAL GRADIENT")
    elif driver == "hessian" and method == "qcisd(t)":
        # contractual_qcvars.append("QCISD(T) TOTAL GRADIENT")
        contractual_qcvars.append("QCISD(T) TOTAL HESSIAN")

    for pv in contractual_qcvars:
        expected = True

        yield (pv, pv, expected)


def contractual_fci(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    f"""Of the list of QCVariables an ideal FCI should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "FCI CORRELATION ENERGY",
        "FCI TOTAL ENERGY",
    ]
    if driver == "gradient" and method == "fci":
        contractual_qcvars.append("FCI TOTAL GRADIENT")
    elif driver == "hessian" and method == "fci":
        # contractual_qcvars.append("FCI TOTAL GRADIENT")
        contractual_qcvars.append("FCI TOTAL HESSIAN")

    for pv in contractual_qcvars:
        expected = True

        yield (pv, pv, expected)


def contractual_remp2(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    f"""Of the list of QCVariables an ideal REMP2 should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "REMP2 CORRELATION ENERGY",
        "REMP2 TOTAL ENERGY",
        "REMP2 SAME-SPIN CORRELATION ENERGY",
        "REMP2 SINGLES ENERGY",
        "REMP2 DOUBLES ENERGY",
        "REMP2 OPPOSITE-SPIN CORRELATION ENERGY",
    ]
    if driver == "gradient" and method == "remp2":
        contractual_qcvars.append("REMP2 TOTAL GRADIENT")
    elif driver == "hessian" and method == "remp2":
        # contractual_qcvars.append("REMP2 TOTAL GRADIENT")
        contractual_qcvars.append("REMP2 TOTAL HESSIAN")

    for pv in contractual_qcvars:
        expected = True
        if (
            ((qc_module == "psi4-occ" and reference == "rhf" and corl_type in ["df", "cd"] and method == "remp2"))
            and pv in ["REMP2 SAME-SPIN CORRELATION ENERGY", "REMP2 OPPOSITE-SPIN CORRELATION ENERGY"]
        ) or (
            (
                qc_module == "psi4-occ"
                and reference in ["rhf", "uhf", "rohf"]
                and corl_type in ["conv", "df", "cd"]
                and method == "oremp2"
            )
            and pv
            in [
                "REMP2 CORRELATION ENERGY",
                "REMP2 TOTAL ENERGY",
                "REMP2 SAME-SPIN CORRELATION ENERGY",
                "REMP2 SINGLES ENERGY",
                "REMP2 DOUBLES ENERGY",
                "REMP2 OPPOSITE-SPIN CORRELATION ENERGY",
            ]
        ):
            expected = False

        yield (pv, pv, expected)


def contractual_lccd(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    f"""Of the list of QCVariables an ideal LCCD should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "LCCD CORRELATION ENERGY",
        "LCCD TOTAL ENERGY",
        "LCCD SAME-SPIN CORRELATION ENERGY",
        "LCCD SINGLES ENERGY",
        "LCCD DOUBLES ENERGY",
        "LCCD OPPOSITE-SPIN CORRELATION ENERGY",
    ]
    if driver == "gradient" and method == "lccd":
        contractual_qcvars.append("LCCD TOTAL GRADIENT")
    elif driver == "hessian" and method == "lccd":
        # contractual_qcvars.append("LCCD TOTAL GRADIENT")
        contractual_qcvars.append("LCCD TOTAL HESSIAN")

    for pv in contractual_qcvars:
        expected = True
        if (
            (
                (
                    (qc_module == "psi4-occ" and reference == "rhf" and corl_type in ["df", "cd"] and method == "lccd")
                    or (qc_module == "cfour-ncc" and reference in ["rhf"] and method == "lccd")
                    or (qc_module == "nwchem-tce" and reference in ["rhf", "uhf"] and method == "lccd")
                    or (qc_module == "gamess" and reference in ["rhf"] and method == "lccd")
                )
                and pv in ["LCCD SAME-SPIN CORRELATION ENERGY", "LCCD OPPOSITE-SPIN CORRELATION ENERGY"]
            )
            or (
                (qc_module == "nwchem-tce" and reference in ["rohf"] and method in ["lccd"])
                and pv
                in [
                    "LCCD SAME-SPIN CORRELATION ENERGY",
                    "LCCD OPPOSITE-SPIN CORRELATION ENERGY",
                    "LCCD SINGLES ENERGY",
                    "LCCD DOUBLES ENERGY",
                ]
            )
            or (
                (
                    qc_module == "psi4-occ"
                    and reference in ["rhf", "uhf", "rohf"]
                    and corl_type in ["conv", "df", "cd"]
                    and method == "olccd"
                )
                and pv
                in [
                    "LCCD CORRELATION ENERGY",
                    "LCCD TOTAL ENERGY",
                    "LCCD SAME-SPIN CORRELATION ENERGY",
                    "LCCD SINGLES ENERGY",
                    "LCCD DOUBLES ENERGY",
                    "LCCD OPPOSITE-SPIN CORRELATION ENERGY",
                ]
            )
        ):
            expected = False

        yield (pv, pv, expected)


def contractual_lccsd(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    f"""Of the list of QCVariables an ideal LCCSD should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "LCCSD CORRELATION ENERGY",
        "LCCSD TOTAL ENERGY",
        "LCCSD SAME-SPIN CORRELATION ENERGY",
        "LCCSD SINGLES ENERGY",
        "LCCSD DOUBLES ENERGY",
        "LCCSD OPPOSITE-SPIN CORRELATION ENERGY",
    ]
    if driver == "gradient" and method == "lccsd":
        contractual_qcvars.append("LCCSD TOTAL GRADIENT")
    elif driver == "hessian" and method == "lccsd":
        # contractual_qcvars.append("LCCSD TOTAL GRADIENT")
        contractual_qcvars.append("LCCSD TOTAL HESSIAN")

    for pv in contractual_qcvars:
        expected = True
        if (
            (
                (
                    (qc_module == "cfour-ncc" and reference in ["rhf"] and method == "lccsd")
                    or (qc_module == "nwchem-tce" and reference in ["rhf", "uhf"] and method == "lccsd")
                )
                and pv in ["LCCSD SAME-SPIN CORRELATION ENERGY", "LCCSD OPPOSITE-SPIN CORRELATION ENERGY"]
            )
            or (
                (qc_module == "cfour-vcc" and reference in ["rohf"] and method in ["lccsd"])
                and pv
                in [
                    "LCCSD SAME-SPIN CORRELATION ENERGY",
                    "LCCSD SINGLES ENERGY",
                    "LCCSD DOUBLES ENERGY",
                ]
            )
            or (
                (qc_module == "nwchem-tce" and reference in ["rohf"] and method in ["lccsd"])
                and pv
                in [
                    "LCCSD SAME-SPIN CORRELATION ENERGY",
                    "LCCSD OPPOSITE-SPIN CORRELATION ENERGY",
                    "LCCSD SINGLES ENERGY",
                    "LCCSD DOUBLES ENERGY",
                ]
            )
        ):
            expected = False

        yield (pv, pv, expected)


def contractual_cepa_pr1_pr(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    f"""Of the list of QCVariables an ideal CEPA(1) should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "CEPA(1) CORRELATION ENERGY",
        "CEPA(1) TOTAL ENERGY",
        "CEPA(1) SAME-SPIN CORRELATION ENERGY",
        "CEPA(1) SINGLES ENERGY",
        "CEPA(1) DOUBLES ENERGY",
        "CEPA(1) OPPOSITE-SPIN CORRELATION ENERGY",
    ]
    if driver == "gradient" and method == "cepa(1)":
        contractual_qcvars.append("CEPA(1) TOTAL GRADIENT")
    elif driver == "hessian" and method == "cepa(1)":
        # contractual_qcvars.append("CEPA(1) TOTAL GRADIENT")
        contractual_qcvars.append("CEPA(1) TOTAL HESSIAN")

    for pv in contractual_qcvars:
        expected = True
        if (False) and pv in ["CEPA(1) SAME-SPIN CORRELATION ENERGY", "CEPA(1) OPPOSITE-SPIN CORRELATION ENERGY"]:
            expected = False

        yield (pv, pv, expected)


def contractual_cepa_pr3_pr(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    f"""Of the list of QCVariables an ideal CEPA(3) should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "CEPA(3) CORRELATION ENERGY",
        "CEPA(3) TOTAL ENERGY",
        "CEPA(3) SAME-SPIN CORRELATION ENERGY",
        "CEPA(3) SINGLES ENERGY",
        "CEPA(3) DOUBLES ENERGY",
        "CEPA(3) OPPOSITE-SPIN CORRELATION ENERGY",
    ]
    if driver == "gradient" and method == "cepa(3)":
        contractual_qcvars.append("CEPA(3) TOTAL GRADIENT")
    elif driver == "hessian" and method == "cepa(3)":
        # contractual_qcvars.append("CEPA(3) TOTAL GRADIENT")
        contractual_qcvars.append("CEPA(3) TOTAL HESSIAN")

    for pv in contractual_qcvars:
        expected = True
        if (False) and pv in ["CEPA(3) SAME-SPIN CORRELATION ENERGY", "CEPA(3) OPPOSITE-SPIN CORRELATION ENERGY"]:
            expected = False

        yield (pv, pv, expected)


def contractual_acpf(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    f"""Of the list of QCVariables an ideal ACPF should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "ACPF CORRELATION ENERGY",
        "ACPF TOTAL ENERGY",
        "ACPF SAME-SPIN CORRELATION ENERGY",
        "ACPF SINGLES ENERGY",
        "ACPF DOUBLES ENERGY",
        "ACPF OPPOSITE-SPIN CORRELATION ENERGY",
    ]
    if driver == "gradient" and method == "acpf":
        contractual_qcvars.append("ACPF TOTAL GRADIENT")
    elif driver == "hessian" and method == "acpf":
        # contractual_qcvars.append("ACPF TOTAL GRADIENT")
        contractual_qcvars.append("ACPF TOTAL HESSIAN")

    for pv in contractual_qcvars:
        expected = True
        if (False) and pv in ["ACPF SAME-SPIN CORRELATION ENERGY", "ACPF OPPOSITE-SPIN CORRELATION ENERGY"]:
            expected = False

        yield (pv, pv, expected)


def contractual_aqcc(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    f"""Of the list of QCVariables an ideal AQCC should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "AQCC CORRELATION ENERGY",
        "AQCC TOTAL ENERGY",
        "AQCC SAME-SPIN CORRELATION ENERGY",
        "AQCC SINGLES ENERGY",
        "AQCC DOUBLES ENERGY",
        "AQCC OPPOSITE-SPIN CORRELATION ENERGY",
    ]
    if driver == "gradient" and method == "aqcc":
        contractual_qcvars.append("AQCC TOTAL GRADIENT")
    elif driver == "hessian" and method == "aqcc":
        # contractual_qcvars.append("AQCC TOTAL GRADIENT")
        contractual_qcvars.append("AQCC TOTAL HESSIAN")

    for pv in contractual_qcvars:
        expected = True
        if (False) and pv in ["AQCC SAME-SPIN CORRELATION ENERGY", "AQCC OPPOSITE-SPIN CORRELATION ENERGY"]:
            expected = False

        yield (pv, pv, expected)


def contractual_ccd(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    """Of the list of QCVariables an ideal CCD should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "CCD CORRELATION ENERGY",
        "CCD TOTAL ENERGY",
        "CCD SAME-SPIN CORRELATION ENERGY",
        "CCD SINGLES ENERGY",
        "CCD DOUBLES ENERGY",
        "CCD OPPOSITE-SPIN CORRELATION ENERGY",
    ]
    if driver == "gradient" and method == "ccd":
        contractual_qcvars.append("CCD TOTAL GRADIENT")
    elif driver == "hessian" and method == "ccd":
        # contractual_qcvars.append("CCD TOTAL GRADIENT")
        contractual_qcvars.append("CCD TOTAL HESSIAN")

    for pv in contractual_qcvars:
        expected = True
        if (
            (
                (qc_module == "cfour-ecc" and reference in ["rhf"] and method == "ccd")
                or (qc_module == "cfour-ncc" and reference in ["rhf"] and method == "ccd")
                or (qc_module == "nwchem-tce" and reference in ["rhf", "uhf"] and method == "ccd")
                or (qc_module == "gamess" and reference in ["rhf"] and method == "ccd")
                or (qc_module == "psi4-occ" and reference in ["rhf", "uhf"] and method == "ccd")
            )
            and pv in ["CCD SAME-SPIN CORRELATION ENERGY", "CCD OPPOSITE-SPIN CORRELATION ENERGY"]
        ) or (
            (
                (qc_module == "cfour-vcc" and reference in ["rohf"] and method in ["ccd"])
                or (qc_module == "nwchem-tce" and reference in ["rohf"] and method in ["ccd"])
            )
            and pv
            in [
                "CCD SAME-SPIN CORRELATION ENERGY",
                "CCD SINGLES ENERGY",
                "CCD DOUBLES ENERGY",
                "CCD OPPOSITE-SPIN CORRELATION ENERGY",
            ]
        ):
            expected = False

        yield (pv, pv, expected)


def contractual_bccd(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    """Of the list of QCVariables an ideal BCCD should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "BCCD CORRELATION ENERGY",
        "BCCD TOTAL ENERGY",
    ]
    if driver == "gradient" and method == "bccd":
        contractual_qcvars.append("BCCD TOTAL GRADIENT")
    elif driver == "hessian" and method == "bccd":
        # contractual_qcvars.append("BCCD TOTAL GRADIENT")
        contractual_qcvars.append("BCCD TOTAL HESSIAN")

    for pv in contractual_qcvars:
        expected = True
        if False:
            expected = False

        yield (pv, pv, expected)


def contractual_cc2(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    """Of the list of QCVariables an ideal CC2 should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "CC2 CORRELATION ENERGY",
        "CC2 TOTAL ENERGY",
    ]
    if driver == "gradient" and method == "cc2":
        contractual_qcvars.append("CC2 TOTAL GRADIENT")
    elif driver == "hessian" and method == "cc2":
        # contractual_qcvars.append("CC2 TOTAL GRADIENT")
        contractual_qcvars.append("CC2 TOTAL HESSIAN")

    for pv in contractual_qcvars:
        expected = True

        yield (pv, pv, expected)


def contractual_ccsd(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    """Of the list of QCVariables an ideal CCSD should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "CCSD CORRELATION ENERGY",
        "CCSD TOTAL ENERGY",
        "CCSD SAME-SPIN CORRELATION ENERGY",
        "CCSD SINGLES ENERGY",
        "CCSD DOUBLES ENERGY",
        "CCSD OPPOSITE-SPIN CORRELATION ENERGY",
    ]
    if driver == "gradient" and method == "ccsd":
        contractual_qcvars.append("CCSD TOTAL GRADIENT")
    elif driver == "hessian" and method == "ccsd":
        # contractual_qcvars.append("CCSD TOTAL GRADIENT")
        contractual_qcvars.append("CCSD TOTAL HESSIAN")

    for pv in contractual_qcvars:
        expected = True
        if (
            (
                (
                    (qc_module == "gamess" and reference == "rhf" and method in ["ccsd", "ccsd+t(ccsd)", "ccsd(t)"])
                    or (
                        qc_module == "nwchem-tce"
                        and reference in ["rhf", "uhf"]
                        and method in ["ccsd", "ccsd+t(ccsd)", "ccsd(t)"]
                    )
                    or (
                        qc_module in ["cfour-ncc", "cfour-ecc"]
                        and reference in ["rhf"]
                        and method in ["ccsd", "ccsd+t(ccsd)", "ccsd(t)", "a-ccsd(t)"]
                    )
                    or (
                        qc_module == "psi4-occ"
                        and reference in ["rhf", "uhf"]
                        and corl_type in ["df", "cd"]
                        and method in ["ccsd", "ccsd(t)", "a-ccsd(t)"]
                    )
                    or (qc_module in ["cfour-vcc"] and reference in ["rhf", "uhf"] and method in ["ccsd+t(ccsd)"])
                    or (
                        qc_module == "psi4-mrcc"
                        and reference in ["rhf", "uhf"]
                        and method in ["ccsd", "ccsd(t)", "a-ccsd(t)"]
                    )
                )
                and pv in ["CCSD SAME-SPIN CORRELATION ENERGY", "CCSD OPPOSITE-SPIN CORRELATION ENERGY"]
            )
            or (
                (qc_module == "cfour-vcc" and reference in ["rohf"] and method in ["ccsd", "ccsd(t)"])
                and pv
                in [
                    "CCSD SAME-SPIN CORRELATION ENERGY",
                    "CCSD SINGLES ENERGY",
                    "CCSD DOUBLES ENERGY",
                ]
            )
            or (
                (qc_module == "cfour-ecc" and reference in ["rohf"] and method in ["ccsd", "ccsd(t)"])
                and pv
                in [
                    "CCSD OPPOSITE-SPIN CORRELATION ENERGY",
                    "CCSD SINGLES ENERGY",
                    "CCSD DOUBLES ENERGY",
                ]
            )
            or (
                (
                    (qc_module == "gamess" and reference in ["rohf"] and method == "ccsd")
                    or (qc_module == "nwchem-tce" and reference in ["rohf"] and method in ["ccsd", "ccsd(t)"])
                    or (
                        qc_module == "psi4-mrcc"
                        and reference in ["rohf"]
                        and method in ["ccsd", "ccsd(t)", "a-ccsd(t)"]
                    )
                )
                and pv
                in [
                    "CCSD SAME-SPIN CORRELATION ENERGY",
                    "CCSD OPPOSITE-SPIN CORRELATION ENERGY",
                    "CCSD SINGLES ENERGY",
                    "CCSD DOUBLES ENERGY",
                ]
            )
            or (
                ((qc_module == "cfour-vcc" and method in ["bccd", "bccd(t)"]))
                and pv
                in [
                    "CCSD CORRELATION ENERGY",
                    "CCSD TOTAL ENERGY",
                    "CCSD SAME-SPIN CORRELATION ENERGY",
                    "CCSD OPPOSITE-SPIN CORRELATION ENERGY",
                    "CCSD SINGLES ENERGY",
                    "CCSD DOUBLES ENERGY",
                ]
            )
        ):
            expected = False

        yield (pv, pv, expected)

    # TODO check CUSTOM SCS-CCSD _absent_


def contractual_ccsdpt_prccsd_pr(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    """Of the list of QCVariables an ideal CCSD+T(CCSD) (aka CCSD[T]) should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "T(CCSD) CORRECTION ENERGY",
        "CCSD+T(CCSD) CORRELATION ENERGY",
        "CCSD+T(CCSD) TOTAL ENERGY",
    ]
    if driver == "gradient" and method == "ccsd+t(ccsd)":
        contractual_qcvars.append("CCSD+T(CCSD) TOTAL GRADIENT")
    elif driver == "hessian" and method == "ccsd+t(ccsd)":
        # contractual_qcvars.append("CCSD+T(CCSD) TOTAL GRADIENT")
        contractual_qcvars.append("CCSD+T(CCSD) TOTAL HESSIAN")

    for pv in contractual_qcvars:
        expected = True
        if False:
            expected = False

        yield (pv, pv, expected)


def contractual_ccsd_prt_pr(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    """Of the list of QCVariables an ideal CCSD(T) should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "(T) CORRECTION ENERGY",
        "CCSD(T) CORRELATION ENERGY",
        "CCSD(T) TOTAL ENERGY",
    ]
    if driver == "gradient" and method == "ccsd(t)":
        contractual_qcvars.append("CCSD(T) TOTAL GRADIENT")
    elif driver == "hessian" and method == "ccsd(t)":
        # contractual_qcvars.append("CCSD(T) TOTAL GRADIENT")
        contractual_qcvars.append("CCSD(T) TOTAL HESSIAN")

    for pv in contractual_qcvars:
        # print("WW", qc_module, driver, reference, method, corl_type, fcae, pv)
        expected = True
        if (
            (qc_module == "psi4-ccenergy" and method == "bccd(t)") or (qc_module == "cfour-vcc" and method == "bccd(t)")
        ) and pv in [
            "(T) CORRECTION ENERGY",
            "CCSD(T) CORRELATION ENERGY",
            "CCSD(T) TOTAL ENERGY",
        ]:
            expected = False

        yield (pv, pv, expected)


def contractual_accsd_prt_pr(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    """Of the list of QCVariables an ideal A-CCSD(T) (aka Lambda-CCSD(T), aka CCSD(aT) should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "A-(T) CORRECTION ENERGY",
        "A-CCSD(T) CORRELATION ENERGY",
        "A-CCSD(T) TOTAL ENERGY",
    ]
    if driver == "gradient" and method == "a-ccsd(t)":
        contractual_qcvars.append("A-CCSD(T) TOTAL GRADIENT")
    elif driver == "hessian" and method == "a-ccsd(t)":
        # contractual_qcvars.append("A-CCSD(T) TOTAL GRADIENT")
        contractual_qcvars.append("A-CCSD(T) TOTAL HESSIAN")

    for pv in contractual_qcvars:
        expected = True
        if False:
            expected = False

        yield (pv, pv, expected)


def contractual_bccd_prt_pr(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    """Of the list of QCVariables an ideal BCCD(T) should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "B(T) CORRECTION ENERGY",
        "BCCD(T) CORRELATION ENERGY",
        "BCCD(T) TOTAL ENERGY",
    ]
    if driver == "gradient" and method == "bccd(t)":
        contractual_qcvars.append("BCCD(T) TOTAL GRADIENT")
    elif driver == "hessian" and method == "bccd(t)":
        # contractual_qcvars.append("BCCD(T) TOTAL GRADIENT")
        contractual_qcvars.append("BCCD(T) TOTAL HESSIAN")

    for pv in contractual_qcvars:
        expected = True
        if False:
            expected = False

        yield (pv, pv, expected)


def contractual_cc3(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    """Of the list of QCVariables an ideal CC3 should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "CC3 CORRELATION ENERGY",
        "CC3 TOTAL ENERGY",
    ]
    if driver == "gradient" and method == "cc3":
        contractual_qcvars.append("CC3 TOTAL GRADIENT")
    elif driver == "hessian" and method == "cc3":
        # contractual_qcvars.append("CC3 TOTAL GRADIENT")
        contractual_qcvars.append("CC3 TOTAL HESSIAN")

    for pv in contractual_qcvars:
        expected = True

        yield (pv, pv, expected)


def contractual_ccsdt(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    f"""Of the list of QCVariables an ideal CCSDT should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "CCSDT CORRELATION ENERGY",
        "CCSDT TOTAL ENERGY",
    ]
    if driver == "gradient" and method == "ccsdt":
        contractual_qcvars.append("CCSDT TOTAL GRADIENT")
    elif driver == "hessian" and method == "ccsdt":
        # contractual_qcvars.append("CCSDT TOTAL GRADIENT")
        contractual_qcvars.append("CCSDT TOTAL HESSIAN")

    for pv in contractual_qcvars:
        expected = True

        yield (pv, pv, expected)


def contractual_ccsdt1a(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    f"""Of the list of QCVariables an ideal CCSDT-1A should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "CCSDT-1A CORRELATION ENERGY",
        "CCSDT-1A TOTAL ENERGY",
    ]
    if driver == "gradient" and method == "ccsdt-1a":
        contractual_qcvars.append("CCSDT-1A TOTAL GRADIENT")
    elif driver == "hessian" and method == "ccsdt-1a":
        # contractual_qcvars.append("CCSDT-1A TOTAL GRADIENT")
        contractual_qcvars.append("CCSDT-1A TOTAL HESSIAN")

    for pv in contractual_qcvars:
        expected = True

        yield (pv, pv, expected)


def contractual_ccsdt1b(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    f"""Of the list of QCVariables an ideal CCSDT-1B should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "CCSDT-1B CORRELATION ENERGY",
        "CCSDT-1B TOTAL ENERGY",
    ]
    if driver == "gradient" and method == "ccsdt-1b":
        contractual_qcvars.append("CCSDT-1B TOTAL GRADIENT")
    elif driver == "hessian" and method == "ccsdt-1b":
        # contractual_qcvars.append("CCSDT-1B TOTAL GRADIENT")
        contractual_qcvars.append("CCSDT-1B TOTAL HESSIAN")

    for pv in contractual_qcvars:
        expected = True

        yield (pv, pv, expected)


def contractual_ccsdt2(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    f"""Of the list of QCVariables an ideal CCSDT-2 should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "CCSDT-2 CORRELATION ENERGY",
        "CCSDT-2 TOTAL ENERGY",
    ]
    if driver == "gradient" and method == "ccsdt-2":
        contractual_qcvars.append("CCSDT-2 TOTAL GRADIENT")
    elif driver == "hessian" and method == "ccsdt-2":
        # contractual_qcvars.append("CCSDT-2 TOTAL GRADIENT")
        contractual_qcvars.append("CCSDT-2 TOTAL HESSIAN")

    for pv in contractual_qcvars:
        expected = True

        yield (pv, pv, expected)


def contractual_ccsdt3(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    f"""Of the list of QCVariables an ideal CCSDT-3 should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "CCSDT-3 CORRELATION ENERGY",
        "CCSDT-3 TOTAL ENERGY",
    ]
    if driver == "gradient" and method == "ccsdt-3":
        contractual_qcvars.append("CCSDT-3 TOTAL GRADIENT")
    elif driver == "hessian" and method == "ccsdt-3":
        # contractual_qcvars.append("CCSDT-3 TOTAL GRADIENT")
        contractual_qcvars.append("CCSDT-3 TOTAL HESSIAN")

    for pv in contractual_qcvars:
        expected = True

        yield (pv, pv, expected)


def contractual_ccsdt_prq_pr(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    """Of the list of QCVariables an ideal CCSDT(Q) should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "(Q) CORRECTION ENERGY",
        "CCSDT(Q) CORRELATION ENERGY",
        "CCSDT(Q) TOTAL ENERGY",
    ]
    if driver == "gradient" and method == "ccsdt(q)":
        contractual_qcvars.append("CCSDT(Q) TOTAL GRADIENT")
    elif driver == "hessian" and method == "ccsdt(q)":
        # contractual_qcvars.append("CCSDT(Q) TOTAL GRADIENT")
        contractual_qcvars.append("CCSDT(Q) TOTAL HESSIAN")

    for pv in contractual_qcvars:
        expected = True

        yield (pv, pv, expected)


def contractual_ccsdtq(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    """Of the list of QCVariables an ideal CCSDTQ should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "CCSDTQ CORRELATION ENERGY",
        "CCSDTQ TOTAL ENERGY",
    ]
    if driver == "gradient" and method == "ccsdtq":
        contractual_qcvars.append("CCSDTQ TOTAL GRADIENT")
    elif driver == "hessian" and method == "ccsdtq":
        # contractual_qcvars.append("CCSDTQ TOTAL GRADIENT")
        contractual_qcvars.append("CCSDTQ TOTAL HESSIAN")

    for pv in contractual_qcvars:
        expected = True

        yield (pv, pv, expected)


def contractual_omp2(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    """Of the list of QCVariables an ideal OMP2 should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "OMP2 CORRELATION ENERGY",
        "OMP2 TOTAL ENERGY",
        "OMP2 REFERENCE CORRECTION ENERGY",
        "OMP2 SAME-SPIN CORRELATION ENERGY",
        "OMP2 OPPOSITE-SPIN CORRELATION ENERGY",
    ]
    if driver == "gradient" and method == "omp2":
        contractual_qcvars.append("OMP2 TOTAL GRADIENT")
    elif driver == "hessian" and method == "omp2":
        # contractual_qcvars.append("OMP2 TOTAL GRADIENT")
        contractual_qcvars.append("OMP2 TOTAL HESSIAN")

    for pv in contractual_qcvars:
        expected = True
        if (
            (qc_module == "psi4-occ" and reference == "rhf" and corl_type in ["df", "cd"] and method == "omp2")
        ) and pv in ["OMP2 SAME-SPIN CORRELATION ENERGY", "OMP2 OPPOSITE-SPIN CORRELATION ENERGY"]:
            expected = False

        yield (pv, pv, expected)


def contractual_omp2p5(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    """Of the list of QCVariables an ideal OMP2.5 should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "OMP2.5 CORRELATION ENERGY",
        "OMP2.5 TOTAL ENERGY",
        "OMP2.5 REFERENCE CORRECTION ENERGY",
        "OMP2.5 SAME-SPIN CORRELATION ENERGY",
        "OMP2.5 OPPOSITE-SPIN CORRELATION ENERGY",
    ]
    if driver == "gradient" and method == "omp2.5":
        contractual_qcvars.append("OMP2.5 TOTAL GRADIENT")
    elif driver == "hessian" and method == "omp2.5":
        # contractual_qcvars.append("OMP2.5 TOTAL GRADIENT")
        contractual_qcvars.append("OMP2.5 TOTAL HESSIAN")

    for pv in contractual_qcvars:
        expected = True
        if (
            (qc_module == "psi4-occ" and reference == "rhf" and corl_type in ["df", "cd"] and method == "omp2.5")
        ) and pv in ["OMP2.5 SAME-SPIN CORRELATION ENERGY", "OMP2.5 OPPOSITE-SPIN CORRELATION ENERGY"]:
            expected = False

        yield (pv, pv, expected)


def contractual_omp3(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    """Of the list of QCVariables an ideal OMP3 should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "OMP3 CORRELATION ENERGY",
        "OMP3 TOTAL ENERGY",
        "OMP3 REFERENCE CORRECTION ENERGY",
        "OMP3 SAME-SPIN CORRELATION ENERGY",
        "OMP3 OPPOSITE-SPIN CORRELATION ENERGY",
    ]
    if driver == "gradient" and method == "omp3":
        contractual_qcvars.append("OMP3 TOTAL GRADIENT")
    elif driver == "hessian" and method == "omp3":
        # contractual_qcvars.append("OMP3 TOTAL GRADIENT")
        contractual_qcvars.append("OMP3 TOTAL HESSIAN")

    for pv in contractual_qcvars:
        expected = True
        if (
            (qc_module == "psi4-occ" and reference == "rhf" and corl_type in ["df", "cd"] and method == "omp3")
        ) and pv in ["OMP3 SAME-SPIN CORRELATION ENERGY", "OMP3 OPPOSITE-SPIN CORRELATION ENERGY"]:
            expected = False

        yield (pv, pv, expected)


def contractual_oremp2(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    """Of the list of QCVariables an ideal OREMP2 should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "OREMP2 CORRELATION ENERGY",
        "OREMP2 TOTAL ENERGY",
        "OREMP2 REFERENCE CORRECTION ENERGY",
        "OREMP2 SAME-SPIN CORRELATION ENERGY",
        "OREMP2 OPPOSITE-SPIN CORRELATION ENERGY",
    ]
    if driver == "gradient" and method == "oremp2":
        contractual_qcvars.append("OREMP2 TOTAL GRADIENT")
    elif driver == "hessian" and method == "oremp2":
        # contractual_qcvars.append("OREMP2 TOTAL GRADIENT")
        contractual_qcvars.append("OREMP2 TOTAL HESSIAN")

    for pv in contractual_qcvars:
        expected = True
        if (
            # usual for oo (qc_module == "psi4-occ" and reference == "rhf" and corl_type in ["df", "cd"] and method == "oremp2")
            (
                qc_module == "psi4-occ"
                and reference in ["rhf", "uhf", "rohf"]
                and corl_type in ["df", "cd"]
                and method == "oremp2"
            )
        ) and pv in ["OREMP2 SAME-SPIN CORRELATION ENERGY", "OREMP2 OPPOSITE-SPIN CORRELATION ENERGY"]:
            expected = False

        yield (pv, pv, expected)


def contractual_olccd(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    """Of the list of QCVariables an ideal OLCCD should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "OLCCD CORRELATION ENERGY",
        "OLCCD TOTAL ENERGY",
        "OLCCD REFERENCE CORRECTION ENERGY",
        "OLCCD SAME-SPIN CORRELATION ENERGY",
        "OLCCD OPPOSITE-SPIN CORRELATION ENERGY",
    ]
    if driver == "gradient" and method == "olccd":
        contractual_qcvars.append("OLCCD TOTAL GRADIENT")
    elif driver == "hessian" and method == "olccd":
        # contractual_qcvars.append("OLCCD TOTAL GRADIENT")
        contractual_qcvars.append("OLCCD TOTAL HESSIAN")

    for pv in contractual_qcvars:
        expected = True
        if (
            (qc_module == "psi4-occ" and reference == "rhf" and corl_type in ["df", "cd"] and method == "olccd")
        ) and pv in ["OLCCD SAME-SPIN CORRELATION ENERGY", "OLCCD OPPOSITE-SPIN CORRELATION ENERGY"]:
            expected = False

        yield (pv, pv, expected)


def contractual_occd(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    """Of the list of QCVariables an ideal OCCD should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "OCCD CORRELATION ENERGY",
        "OCCD TOTAL ENERGY",
        "OCCD REFERENCE CORRECTION ENERGY",
        "OCCD SAME-SPIN CORRELATION ENERGY",
        "OCCD OPPOSITE-SPIN CORRELATION ENERGY",
    ]
    if driver == "gradient" and method == "occd":
        contractual_qcvars.append("OCCD TOTAL GRADIENT")
    elif driver == "hessian" and method == "occd":
        # contractual_qcvars.append("OCCD TOTAL GRADIENT")
        contractual_qcvars.append("OCCD TOTAL HESSIAN")

    for pv in contractual_qcvars:
        expected = True
        if (
            (
                qc_module == "psi4-occ"
                and reference in ["rhf", "uhf", "rohf"]
                and corl_type in ["df", "cd"]
                and method in ["occd", "occd(t)", "a-occd(t)"]
            )
        ) and pv in ["OCCD SAME-SPIN CORRELATION ENERGY", "OCCD OPPOSITE-SPIN CORRELATION ENERGY"]:
            expected = False

        yield (pv, pv, expected)


def contractual_occd_prt_pr(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    """Of the list of QCVariables an ideal OCCD(T) should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "O(T) CORRECTION ENERGY",
        "OCCD(T) CORRELATION ENERGY",
        "OCCD(T) TOTAL ENERGY",
    ]
    if driver == "gradient" and method == "occd(t)":
        contractual_qcvars.append("OCCD(T) TOTAL GRADIENT")
    elif driver == "hessian" and method == "occd(t)":
        # contractual_qcvars.append("OCCD(T) TOTAL GRADIENT")
        contractual_qcvars.append("OCCD(T) TOTAL HESSIAN")

    for pv in contractual_qcvars:
        expected = True
        if False:
            expected = False

        yield (pv, pv, expected)


def contractual_aoccd_prt_pr(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    """Of the list of QCVariables an ideal A-OCCD(T) (aka Lambda-OCCD(T), aka OCCD(aT) should produce, returns whether or
    not each is expected, given the calculation circumstances (like QC program).

    {_contractual_docstring}
    """
    contractual_qcvars = [
        "HF TOTAL ENERGY",
        "A-O(T) CORRECTION ENERGY",
        "A-OCCD(T) CORRELATION ENERGY",
        "A-OCCD(T) TOTAL ENERGY",
    ]
    if driver == "gradient" and method == "a-occd(t)":
        contractual_qcvars.append("A-OCCD(T) TOTAL GRADIENT")
    elif driver == "hessian" and method == "a-occd(t)":
        # contractual_qcvars.append("A-OCCD(T) TOTAL GRADIENT")
        contractual_qcvars.append("A-OCCD(T) TOTAL HESSIAN")

    for pv in contractual_qcvars:
        expected = True
        if False:
            expected = False

        yield (pv, pv, expected)


def contractual_dft_current(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    """Given the target DFT method, returns the CURRENT QCVariables that should be produced.

    {_contractual_docstring}
    """
    contractual_qcvars = [
        (f"{method.upper()} TOTAL ENERGY", "DFT TOTAL ENERGY"),
        (f"{method.upper()} TOTAL ENERGY", "CURRENT REFERENCE ENERGY"),
        (f"{method.upper()} TOTAL ENERGY", "CURRENT ENERGY"),
        # assert missing
        ("HF TOTAL ENERGY", "HF TOTAL ENERGY"),
    ]
    if driver == "gradient":
        contractual_qcvars.append((f"{method.upper()} TOTAL GRADIENT", "CURRENT GRADIENT"))
    elif driver == "hessian":
        # contractual_qcvars.append((f"{method.upper()} TOTAL GRADIENT", "CURRENT GRADIENT"))
        contractual_qcvars.append((f"{method.upper()} TOTAL HESSIAN", "CURRENT HESSIAN"))

    for rpv, pv in contractual_qcvars:
        expected = True
        if True and rpv in [  # assert canonical wfn quantities missing in DFT
            "HF TOTAL ENERGY",
        ]:
            expected = False

        yield (rpv, pv, expected)


def contractual_dhdft_current(
    qc_module: str, driver: str, reference: str, method: str, corl_type: str, fcae: str, sdsc: str
) -> Tuple[str, str, bool]:
    """Given the target DFT method, returns the CURRENT QCVariables that should be produced.

    {_contractual_docstring}
    """
    contractual_qcvars = [
        (f"{method.upper()} TOTAL ENERGY", "DFT TOTAL ENERGY"),
        (f"{method.upper()} FUNCTIONAL TOTAL ENERGY", "CURRENT REFERENCE ENERGY"),
        (f"{method.upper()} TOTAL ENERGY", "CURRENT ENERGY"),
        # assert missing
        ("HF TOTAL ENERGY", "HF TOTAL ENERGY"),
        ("MP2 CORRELATION ENERGY", "MP2 CORRELATION ENERGY"),
        ("MP2 TOTAL ENERGY", "MP2 TOTAL ENERGY"),
        ("MP2 SAME-SPIN CORRELATION ENERGY", "MP2 SAME-SPIN CORRELATION ENERGY"),
        ("MP2 SINGLES ENERGY", "MP2 SINGLES ENERGY"),
        ("MP2 DOUBLES ENERGY", "MP2 DOUBLES ENERGY"),
        ("MP2 OPPOSITE-SPIN CORRELATION ENERGY", "MP2 OPPOSITE-SPIN CORRELATION ENERGY"),
    ]
    if driver == "gradient":
        contractual_qcvars.append((f"{method.upper()} TOTAL GRADIENT", "CURRENT GRADIENT"))
    elif driver == "hessian":
        # contractual_qcvars.append((f"{method.upper()} TOTAL GRADIENT", "CURRENT GRADIENT"))
        contractual_qcvars.append((f"{method.upper()} TOTAL HESSIAN", "CURRENT HESSIAN"))

    for rpv, pv in contractual_qcvars:
        expected = True
        if True and rpv in [  # assert canonical wfn quantities missing in DFT
            "HF TOTAL ENERGY",
            "MP2 CORRELATION ENERGY",
            "MP2 TOTAL ENERGY",
            "MP2 SAME-SPIN CORRELATION ENERGY",
            "MP2 SINGLES ENERGY",
            "MP2 DOUBLES ENERGY",
            "MP2 OPPOSITE-SPIN CORRELATION ENERGY",
        ]:
            expected = False

        yield (rpv, pv, expected)



import sys
import traceback
from typing import Any, Dict, List

import numpy as np
from qcelemental.models import BasisSet


def cca_ao_order_spherical(max_angular_momentum: int) -> Dict[int, List[int]]:
    ao_order = {}
    for ang_mom in range(max_angular_momentum):
        ao_order[ang_mom] = [x for x in range(-1 * ang_mom, ang_mom + 1)]
    return ao_order


def get_ao_conversion(new_ao_order: Dict[int, List[int]], old_ao_order: Dict[int, List[int]]) -> Dict[int, List[int]]:
    """
    This function determines the conversion of AO ordering for each angular momentum between the CCA standard and the
    provided program harness.
    """

    to_new_order = {}
    for ang_mom in old_ao_order.keys():
        old_order = old_ao_order[ang_mom]
        new_order = new_ao_order[ang_mom]
        sort_indices = []
        for element in new_order:
            sort_indices.append(old_order.index(element))

        to_new_order[ang_mom] = sort_indices

    return to_new_order


def reorder_column_ao_indices(
    matrix: np.ndarray, basis: BasisSet, to_new_ao_order: Dict[str, Dict[int, List[int]]]
) -> np.ndarray:
    """
    Reorder the column atomic orbital (AO) indices of matrix.
    """

    num_cols = len(matrix[0])
    num_rows = len(matrix)
    new_matrix = np.zeros([num_rows, num_cols])
    for row_index in range(0, num_rows):
        ao_shift = 0
        for atom in basis.atom_map:
            for electron_shell in basis.center_data[atom].electron_shells:
                to_cca_indices = to_new_ao_order[electron_shell.harmonic_type][electron_shell.angular_momentum[0]]
                for bas_index in range(0, len(to_cca_indices)):
                    new_matrix[row_index][ao_shift + to_cca_indices[bas_index]] = matrix[row_index][
                        ao_shift + bas_index
                    ]
                ao_shift += len(to_cca_indices)

    return new_matrix


def reorder_row_and_column_ao_indices(
    matrix: np.ndarray, basis: BasisSet, to_new_ao_order: Dict[str, Dict[int, List[int]]]
) -> np.ndarray:
    """
    Reorder both row and column atomic orbital (AO) indices of matrix. (e.g. Fock, density)
    """

    col_reordered_matrix = reorder_column_ao_indices(matrix, basis, to_new_ao_order)
    new_matrix = reorder_column_ao_indices(col_reordered_matrix.transpose(), basis, to_new_ao_order)

    return new_matrix


def mill_qcvars(mill: "AlignmentMill", qcvars: Dict[str, Any]) -> Dict[str, Any]:
    """Apply translation, rotation, atom shuffle defined in ``mill`` to the nonscalar quantities in ``qcvars``."""

    milled = {}
    for k, v in qcvars.items():
        if isinstance(v, str) and v == "KnownMissing":
            milled[k] = v
        elif k.endswith("GRADIENT"):
            milled[k] = mill.align_gradient(v)
        elif k.endswith("HESSIAN"):
            milled[k] = mill.align_hessian(v)
        else:
            milled[k] = v

    return milled


def error_stamp(stdin: str = "", stdout: str = "", stderr: str = "", tb: str = None) -> str:
    """Return all useful information in error string."""

    if not tb:
        tb = traceback.format_exception(*sys.exc_info())
    return "INPUT:\n" + stdin + "STDOUT:\n" + stdout + "\nSTDERR:\n" + stderr + "\nTRACEBACK:\n" + "".join(tb)



"""
Calls the TeraChem executable.
"""

import re
from typing import Any, Dict, Optional

from qcelemental.models import AtomicResult, FailedOperation
from qcelemental.molparse.regex import DECIMAL, NUMBER
from qcelemental.util import parse_version, safe_version, which

from qcengine import util as uti

from ..exceptions import UnknownError
from ..util import popen
from .model import ProgramHarness


class TeraChemHarness(ProgramHarness):

    _defaults = {
        "name": "TeraChem",
        "scratch": True,
        "thread_safe": False,
        "thread_parallel": True,
        "node_parallel": False,
        "managed_memory": True,
    }
    version_cache: Dict[str, str] = {}

    class Config(ProgramHarness.Config):
        pass

    @staticmethod
    def found(raise_error: bool = False) -> bool:
        return which(
            "terachem",
            return_bool=True,
            raise_error=raise_error,
            raise_msg="Please install via http://www.petachem.com/index.html",
        )

    def get_version(self) -> str:
        self.found(raise_error=True)

        which_prog = which("terachem")
        if which_prog not in self.version_cache:
            with popen([which_prog, "--version"]) as exc:
                exc["proc"].wait(timeout=5)
            mobj = re.search(NUMBER, exc["stdout"], re.VERBOSE)
            version = mobj.group(0)
            self.version_cache[which_prog] = safe_version(version)

        return self.version_cache[which_prog]

    def compute(self, input_data: "AtomicInput", config: "TaskConfig") -> "AtomicResult":
        """
        Run TeraChem
        """
        self.found(raise_error=True)

        # Check TeraChem version
        if parse_version(self.get_version()) < parse_version("1.5"):
            raise TypeError("TeraChem version '{}' not understood".format(self.get_version()))

        # Setup the job
        job_inputs = self.build_input(input_data, config)
        # Run terachem
        exe_outputs = self.execute(job_inputs, extra_outfiles=input_data.extras)
        exe_success, proc = exe_outputs
        # Determine whether the calculation succeeded
        output_data = {}
        if not exe_success:
            output_data["success"] = False
            output_data["error"] = {"error_type": "unknown_error", "error_message": proc["stderr"]}
            return FailedOperation(
                success=output_data.pop("success", False), error=output_data.pop("error"), input_data=output_data
            )

        # If execution succeeded, collect results
        result = self.parse_output(proc["outfiles"], input_data)
        return result

    def build_input(
        self, input_model: "AtomicInput", config: "TaskConfig", template: Optional[str] = None
    ) -> Dict[str, Any]:
        # Write the geom xyz file with unit au
        xyz_file = input_model.molecule.to_string(dtype="terachem", units="Bohr")

        # Write input file
        input_file = []
        input_file.append("# molecule definition")
        input_file.append("units bohr")
        input_file.append("charge " + str(int(input_model.molecule.molecular_charge)))
        input_file.append("spinmult " + str(input_model.molecule.molecular_multiplicity))
        input_file.append("coordinates geometry.xyz")

        input_file.append("\n# model")
        input_file.append("basis " + str(input_model.model.basis))
        input_file.append("method " + str(input_model.model.method))

        input_file.append("\n# driver")
        input_file.append("run " + input_model.driver)

        input_file.append("\n# keywords")
        for k, v in input_model.keywords.items():
            input_file.append("{} {}".format(k, v))

        input_file = "\n".join(input_file)

        return {
            "commands": ["terachem", "tc.in"],
            "infiles": {"tc.in": input_file, "geometry.xyz": xyz_file},
            "scratch_directory": config.scratch_directory,
            "input_result": input_model.copy(deep=True),
        }

    def parse_output(self, outfiles: Dict[str, str], input_model: "AtomicInput") -> "AtomicResult":
        output_data = {}
        properties = {}

        # Parse the output file, collect properties and gradient
        output_lines = outfiles["tc.out"].split("\n")
        gradients = []
        natom = 0
        line_final_energy = -1
        line_scf_header = -1
        for idx, line in enumerate(output_lines):
            if "FINAL ENERGY" in line:
                properties["scf_total_energy"] = float(line.strip("\n").split()[2])
                line_final_energy = idx
            elif "Start SCF Iterations" in line:
                line_scf_header = idx
            elif "Total atoms" in line:
                natom = int(line.split()[-1])
            elif "DIPOLE MOMENT" in line:
                newline = line.replace(",", "").replace("}", "").replace("{", "")
                properties["scf_dipole_moment"] = [float(x) for x in newline.split()[2:5]]
            elif "Nuclear repulsion energy" in line:
                properties["nuclear_repulsion_energy"] = float(line.split()[-2])
            elif "Gradient units are Hartree/Bohr" in line:
                # Gradient is stored as (dE/dx1,dE/dy1,dE/dz1,dE/dx2,dE/dy2,...)
                for i in range(idx + 3, idx + 3 + natom):
                    grad = output_lines[i].strip("\n").split()
                    for x in grad:
                        gradients.append(float(x))

        last_scf_line = ""
        for idx in reversed(range(line_scf_header, line_final_energy)):
            mobj = re.search(
                r"^\s*\d+\s+" + DECIMAL + r"\s+" + DECIMAL + r"\s+" + DECIMAL + r"\s+" + DECIMAL,
                output_lines[idx],
                re.VERBOSE,
            )
            if mobj:
                last_scf_line = output_lines[idx]
                break

        if len(last_scf_line) > 0:
            properties["scf_iterations"] = int(last_scf_line.split()[0])
            if "XC Energy" in output_lines:
                properties["scf_xc_energy"] = float(last_scf_line.split()[4])
        else:
            raise UnknownError("SCF iteration lines not found in TeraChem output")

        if len(gradients) > 0:
            output_data["return_result"] = gradients

        # Commented out the properties currently not supported by QCSchema
        # properites["spin_S2"] = 1 # calculated S(S+1)
        #   elif "SPIN S-SQUARED" in line:
        #       properties["spin_S2"] = float(line.strip('\n').split()[2])
        # Parse files in scratch folder
        # properties["atomic_charge"] = []
        # atomic_charge_lines =  open(outfiles["charge.xls"]).readlines()
        # for line in atomic_charge_lines:
        #    properties["atomic_charge"].append(line.strip('\n').split()[-1])

        if "return_result" not in output_data:
            if "scf_total_energy" in properties:
                output_data["return_result"] = properties["scf_total_energy"]
            else:
                raise KeyError("Could not find SCF total energy")

        output_data["properties"] = properties

        output_data["schema_name"] = "qcschema_output"
        output_data["stdout"] = outfiles["tc.out"]
        # TODO Should only return True if TeraChem calculation terminated properly
        output_data["success"] = True

        # return extra files requested by user as extras
        for extra in input_model.extras.keys():
            input_model.extras[extra] = outfiles[extra]

        return AtomicResult(**{**input_model.dict(), **output_data})

    def execute(self, inputs, extra_outfiles=None, extra_commands=None, scratch_name=None, timeout=None):
        binaries = []
        for filename in extra_outfiles:
            if filename in ["scr/ca0", "scr/cb0", "scr/c0"]:
                binaries.append(filename)
        exe_success, proc = uti.execute(
            inputs["commands"],
            infiles=inputs["infiles"],
            outfiles=extra_outfiles,
            as_binary=binaries,
            scratch_directory=inputs["scratch_directory"],
            timeout=timeout,
        )
        proc["outfiles"]["tc.out"] = proc["stdout"]
        return exe_success, proc



"""Tests for NWChem functionality"""
import numpy as np
import pytest
import qcelemental as qcel
from qcelemental.testing import compare_values

import qcengine as qcng
from qcengine.testing import using

# Molecule where autoz fails
_auto_z_problem = xyz = """C                    15.204188380000    -3.519180270000   -10.798726560000
C                    15.097645630000    -2.650246400000    -8.505033680000
C                    14.976892130000    -1.867030510000    -6.378827230000
C                    14.827093540000    -1.105281910000    -4.239222620000
C                    14.635399720000    -0.362921900000    -2.101035300000
N                    14.357893440000     0.315376390000    -0.000359050000
C                    16.031227020000     2.397930170000     0.801905280000
C                    15.337300690000     2.925144860000     3.538890120000
C                    14.742074760000     3.377413020000     5.886685850000
C                    14.048167320000     3.904627710000     8.623670690000
N                    12.688887320000     2.073331920000     9.820037400000
C                    14.474640720000     6.115210430000     9.591909660000
C                    14.946826580000     8.308766730000    10.603064320000
C                    17.506555110000     9.181593430000    10.935939570000
C                    18.224273090000    10.758664370000    12.778309160000
O                    20.684526430000    11.555996510000    13.018002020000
C                    20.952489590000    13.130270650000    14.964495520000
C                    21.186059740000    14.590046290000    16.687812370000
C                    21.422747940000    16.379824800000    18.663483240000
C                    22.929993500000    15.902177630000    20.539319870000
C                    24.426014080000    15.474645990000    22.431805000000
N                    24.775991360000    13.082933010000    23.323528960000
O                    18.709082330000     1.890803270000     0.670399240000
O                    20.008439110000     4.184817400000     1.512272230000
H                    16.847872170000    -3.183338140000   -11.974003930000
H                    13.649907540000    -4.604695650000   -11.575422900000
H                    15.665451630000     4.107792170000    -0.318135390000
H                    13.044004660000     0.318248780000     9.124920540000
H                    12.835662350000     2.164511200000    11.730569410000
H                    13.401843200000     9.476598570000    11.281381510000
H                    18.900965120000     8.515559460000     9.580741380000
H                    16.945759980000    11.469295880000    14.211420760000
H                    20.389540180000    18.145263640000    18.570527610000
H                    25.288598470000    16.991453560000    23.511896860000
H                    24.436823320000    11.711936710000    22.016121940000
H                    26.442597520000    12.834282850000    24.246149950000
H                    20.850425490000     3.414376060000     2.960577230000"""


@pytest.fixture
def nh2():
    smol = """
 # R=1.008 #A=105.0
 0 2
 N   0.000000000000000   0.000000000000000  -0.145912918634892
 H   0.000000000000000  -1.511214298139000   1.013682596946108
 H   0.000000000000000   1.511214298139000   1.013682596946108
 units au
 symmetry c1
"""
    return qcel.models.Molecule.from_data(smol)


@using("nwchem")
def test_b3lyp(nh2):
    # Run NH2
    resi = {"molecule": nh2, "driver": "energy", "model": {"method": "b3lyp", "basis": "3-21g"}}
    res = qcng.compute(resi, "nwchem", raise_error=True, return_dict=True)

    # Make sure the calculation completed successfully
    assert compare_values(-55.554037, res["return_result"], atol=1e-3)
    assert res["driver"] == "energy"
    assert "provenance" in res
    assert res["success"] is True

    # Check the other status information
    assert res["extras"]["qcvars"]["N ALPHA ELECTRONS"] == "5"
    assert res["extras"]["qcvars"]["N ATOMS"] == "3"
    assert res["extras"]["qcvars"]["N BASIS FUNCTIONS"] == "13"

    # Make sure the properties parsed correctly
    assert compare_values(-55.554037, res["properties"]["return_energy"], atol=1e-3)
    assert res["properties"]["calcinfo_natom"] == 3
    assert res["properties"]["calcinfo_nalpha"] == 5
    assert res["properties"]["calcinfo_nbasis"] == 13


@using("nwchem")
def test_hess(nh2):
    resi = {"molecule": nh2, "driver": "hessian", "model": {"method": "b3lyp", "basis": "3-21g"}}
    res = qcng.compute(resi, "nwchem", raise_error=True, return_dict=False)
    assert compare_values(-3.5980754370e-02, res.return_result[0, 0], atol=1e-3)
    assert compare_values(0, res.return_result[1, 0], atol=1e-3)
    assert compare_values(0.018208307756, res.return_result[3, 0], atol=1e-3)
    assert np.allclose(res.return_result, res.return_result.T, atol=1e-8)  # Should be symmetric about diagonal

    # Test that the Hessian changes with rotation, but that its determinants remain the same
    shifted_nh2, _ = nh2.scramble(do_shift=False, do_mirror=False, do_rotate=True, do_resort=False)

    resi["molecule"] = shifted_nh2
    res_shifted = qcng.compute(resi, "nwchem", raise_error=True, return_dict=False)
    assert not np.allclose(res.return_result, res_shifted.return_result, atol=1e-8)
    assert np.isclose(np.linalg.det(res.return_result), np.linalg.det(res_shifted.return_result))


@using("nwchem")
def test_gradient(nh2):
    resi = {
        "molecule": nh2,
        "driver": "gradient",
        "model": {"method": "b3lyp", "basis": "3-21g"},
        "keywords": {"dft__convergence__gradient": "1e-6"},
    }
    res = qcng.compute(resi, "nwchem", raise_error=True, return_dict=True)
    assert compare_values(4.22418267e-2, res["return_result"][2], atol=1e-7)  # Beyond accuracy of NWChem stdout

    # Rotate the molecule and verify that the gradient changes
    shifted_nh2, _ = nh2.scramble(do_shift=False, do_mirror=False, do_rotate=True, do_resort=False)

    resi["molecule"] = shifted_nh2
    res_shifted = qcng.compute(resi, "nwchem", raise_error=True, return_dict=True)

    assert not compare_values(4.22418267e-2, res_shifted["return_result"][2], atol=1e-7)

    # Make sure the two matrices still have the same determinant and norms, as they are just rotations of each other
    #  I am leveraging the fact that the gradients are square, 3x3 matrices just by happenstance of the
    #  test molecule having 3 atoms
    orig_grads = np.reshape(res["return_result"], (-1, 3))
    shif_grads = np.reshape(res_shifted["return_result"], (-1, 3))

    # Test that the magnitude of forces are the same
    assert np.allclose(np.linalg.norm(orig_grads, ord=2, axis=1), np.linalg.norm(shif_grads, ord=2, axis=1))

    # Test that the determinants are the same
    orig_det = np.linalg.det(orig_grads)
    shif_det = np.linalg.det(shif_grads)

    assert np.allclose(orig_det, shif_det)


@pytest.fixture
def h20():
    water = """
-1 2
O 0 0 0
H 0 0 1
H 0 1 0
    """
    return qcel.models.Molecule.from_data(water)


@using("nwchem")
def test_dipole(h20):
    # Run NH2
    resi = {
        "molecule": h20,
        "driver": "properties",
        "model": {"method": "dft", "basis": "3-21g"},
        "keywords": {"dft__xc": "b3lyp", "property__dipole": True},
    }
    res = qcng.compute(resi, "nwchem", raise_error=True, return_dict=True)

    # Make sure the calculation completed successfully
    assert compare_values(-75.764944, res["return_result"], atol=1e-3)
    assert res["driver"] == "properties"
    assert "provenance" in res
    assert res["success"] is True

    # Check the other status information
    assert res["extras"]["qcvars"]["N ALPHA ELECTRONS"] == "6"
    assert res["extras"]["qcvars"]["N ATOMS"] == "3"
    assert res["extras"]["qcvars"]["N BASIS FUNCTIONS"] == "13"

    # Make sure the properties parsed correctly
    assert compare_values(-75.764944, res["properties"]["return_energy"], atol=1e-3)
    assert res["properties"]["calcinfo_natom"] == 3
    assert res["properties"]["calcinfo_nalpha"] == 6
    assert res["properties"]["calcinfo_nbasis"] == 13
    # Make sure Dipole Moment and center of charge parsed correctly
    assert compare_values(0.272949872, float(res["extras"]["qcvars"]["TOTAL DIPOLE MOMENT"]), atol=1e-5)
    assert compare_values(-0.00, float(res["extras"]["qcvars"]["DIPOLE MOMENT"][0]), atol=1e-3)
    assert compare_values(-0.00, float(res["extras"]["qcvars"]["DIPOLE MOMENT"][1]), atol=1e-3)
    assert compare_values(-0.272949872, float(res["extras"]["qcvars"]["DIPOLE MOMENT"][2]), atol=1e-5)


@pytest.fixture
def h20v2():
    water = """
O 0 0 0
H 0 0 1
H 0 1 0
    """
    return qcel.models.Molecule.from_data(water)


@using("nwchem")
def test_homo_lumo(h20v2):
    # Run NH2
    resi = {
        "molecule": h20v2,
        "driver": "energy",
        "model": {"method": "dft", "basis": "3-21g"},
        "keywords": {"dft__xc": "b3lyp"},
    }
    res = qcng.compute(resi, "nwchem", raise_error=True, return_dict=True)

    # Make sure the calculation completed successfully
    assert compare_values(-75.968095, res["return_result"], atol=1e-3)
    assert res["driver"] == "energy"
    assert "provenance" in res
    assert res["success"] is True

    # Check the other status information
    assert res["extras"]["qcvars"]["N ALPHA ELECTRONS"] == "5"
    assert res["extras"]["qcvars"]["N ATOMS"] == "3"
    assert res["extras"]["qcvars"]["N BASIS FUNCTIONS"] == "13"

    # Make sure the properties parsed correctly
    assert compare_values(-75.968095, res["properties"]["return_energy"], atol=1e-3)
    assert res["properties"]["calcinfo_natom"] == 3
    assert res["properties"]["calcinfo_nalpha"] == 5
    assert res["properties"]["calcinfo_nbasis"] == 13
    # Make sure Dipole Moment and center of charge parsed correctly
    assert compare_values(-0.2636515, float(res["extras"]["qcvars"]["HOMO"][0]), atol=1e-5)
    assert compare_values(0.08207131, float(res["extras"]["qcvars"]["LUMO"][0]), atol=1e-5)


@using("nwchem")
def test_geometry_bug():
    """Make sure that the harvester does not crash if NWChem's autosym moves atoms too far"""

    # Example molecule that has an RMSD of 2e-4 after NWChem symmetrizes the coordinates
    xyz = """6
Properties=species:S:1:pos:R:3 unique_id=2da214efcf4e4c277fabc5b2b6ca6f32 pbc="F F F"
C      -0.00828817       1.39046978      -0.00560069
O      -0.00797038      -0.02504537       0.02030606
H       1.00658338       1.81556366       0.00348335
H      -0.54657475       1.79916975      -0.87390126
H      -0.52288871       1.72555240       0.89907326
H       0.44142019      -0.33354425      -0.77152059"""
    mol = qcel.models.Molecule.from_data(xyz)
    qcng.compute(
        {"molecule": mol, "model": {"method": "b3lyp", "basis": "6-31g"}, "driver": "gradient"},
        "nwchem",
        raise_error=True,
    )


@using("nwchem")
def test_autoz_error():
    """Test ability to turn off autoz"""
    # Large molecule that leads to an AutoZ error
    mol = qcel.models.Molecule.from_data(_auto_z_problem)
    result = qcng.compute(
        {
            "molecule": mol,
            "model": {"method": "hf", "basis": "sto-3g"},
            "driver": "energy",
            "protocols": {"error_correction": {"default_policy": False}},
        },  # Turn off error correction
        "nwchem",
        raise_error=False,
    )

    assert not result.success
    assert "Error when generating redundant atomic coordinates" in result.error.error_message

    # Turn off autoz
    result = qcng.compute(
        {
            "molecule": mol,
            "model": {"method": "hf", "basis": "sto-3g"},
            "driver": "energy",
            "keywords": {"geometry__noautoz": True},
        },
        "nwchem",
        raise_error=False,
    )

    # Ok if it crashes for other reasons
    assert "Error when generating redundant atomic coordinates" not in result.error.error_message


@using("nwchem")
def test_autoz_error_correction():
    """See if error correction for autoz works"""

    # Large molecule that leads to an AutoZ error
    mol = qcel.models.Molecule.from_data(_auto_z_problem)
    result = qcng.compute(
        {
            "molecule": mol,
            "model": {"method": "hf", "basis": "sto-3g"},
            "driver": "energy",
            "keywords": {"scf__maxiter": 250, "scf__thresh": 1e-1},
        },
        "nwchem",
        raise_error=True,
    )

    assert result.success
    assert "geom_binvr" in result.extras["observed_errors"]
    assert result.extras["observed_errors"]["geom_binvr"]["keyword_updates"] == {"geometry__noautoz": True}


@pytest.mark.parametrize(
    "method, keyword, init_iters, use_tce",
    [
        ["b3lyp", "dft__maxiter", 4, False],
        ["b3lyp", "dft__iterations", 4, False],
        ["hf", "scf__maxiter", 2, False],
        ["mp2", "scf__maxiter", 2, False],
        ["mp2", "scf__maxiter", 2, True],
        ["ccsd", "tce__maxiter", 8, True],
        ["ccsd", "ccsd__maxiter", 4, False],
    ],
)
@using("nwchem")
def test_conv_threshold(h20v2, method, keyword, init_iters, use_tce):
    result = qcng.compute(
        {
            "molecule": h20v2,
            "model": {"method": method, "basis": "sto-3g"},
            "driver": "energy",
            "keywords": {
                keyword: init_iters,
                "qc_module": use_tce,
                "scf__uhf": True,
            },  # UHF needed for SCF test
        },
        "nwchem",
        raise_error=True,
    )

    assert result.success
    assert "convergence_failed" in result.extras["observed_errors"]
    assert result.extras["observed_errors"]["convergence_failed"]["keyword_updates"] == {keyword: init_iters * 4}


@using("nwchem")
def test_restart(nh2, tmpdir):
    # Create a molecule that takes 5-8 steps for NWChem to relax it,
    #  but only run the relaxation for 4 steps
    resi = {
        "molecule": nh2,
        "driver": "gradient",
        "model": {"method": "b3lyp", "basis": "3-21g"},
        "keywords": {"dft__convergence__gradient": "1e-6", "dft__iterations": 4},
        "protocols": {"error_correction": {"default_policy": False}},
        "extras": {"allow_restarts": True},
    }

    # Run once: It should fail to converge
    local_options = {"scratch_messy": True, "scratch_directory": str(tmpdir)}
    result = qcng.compute(
        resi,
        "nwchem",
        local_options=local_options,
        raise_error=False,
    )
    assert not result.success
    assert "computation failed to converge" in str(result.error)

    # Run again: It should converge only if we start from the last geometry
    result = qcng.compute(
        resi,
        "nwchem",
        local_options=local_options,
        raise_error=False,
    )
    assert result.success



import pytest
import qcelemental as qcel
from qcelemental.testing import compare_recursive

import qcengine as qcng
from qcengine.testing import qcengine_records, using

# Prep globals
terachem_info = qcengine_records("terachem")


@pytest.mark.parametrize("test_case", terachem_info.list_test_cases())
def test_terachem_output_parser(test_case):
    # Get output file data
    data = terachem_info.get_test_data(test_case)
    inp = qcel.models.AtomicInput.parse_raw(data["input.json"])

    output = qcng.get_program("terachem", check=False).parse_output(data, inp).dict()
    output_ref = qcel.models.AtomicResult.parse_raw(data["output.json"]).dict()

    # Forgiving molecule since it is now sparse
    assert compare_recursive(output_ref, output, forgive={"stdout", "provenance", "molecule"})


@pytest.mark.parametrize("test_case", terachem_info.list_test_cases())
def test_terachem_input_formatter(test_case):
    # Get input file data
    data = terachem_info.get_test_data(test_case)
    inp = qcel.models.AtomicInput.parse_raw(data["input.json"])

    # TODO add actual comparison of generated input file
    input_file = qcng.get_program("terachem", check=False).build_input(inp, qcng.get_config())
    assert input_file.keys() >= {"commands", "infiles"}


@using("terachem")
@pytest.mark.parametrize("test_case", terachem_info.list_test_cases())
def test_terachem_executor(test_case):
    # Get input file data
    data = terachem_info.get_test_data(test_case)
    inp = qcel.models.AtomicInput.parse_raw(data["input.json"])

    # Run Terachem
    result = qcng.compute(inp, "terachem")
    # result = qcng.get_program('terachem').compute(inp, qcng.get_config())
    assert result.success is True

    # Get output file data
    output_ref = qcel.models.AtomicResult.parse_raw(data["output.json"])

    atol = 1e-6
    if result.driver == "gradient":
        atol = 1e-3
    assert compare_recursive(output_ref.return_result, result.return_result, atol=atol)



"""Compute quantum chemistry using Iowa State's GAMESS executable."""

import logging
import pprint
import re
from decimal import Decimal
from typing import Dict, Optional, Tuple

import numpy as np
import qcelemental as qcel
from qcelemental.models import Molecule
from qcelemental.molparse import regex

from ..util import PreservingDict, load_hessian

pp = pprint.PrettyPrinter(width=120, compact=True, indent=1)
logger = logging.getLogger(__name__)


def harvest(
    in_mol: Molecule, method: str, gamessout: str, **outfiles
) -> Tuple[PreservingDict, Optional[np.ndarray], Optional[np.ndarray], Molecule]:
    """Parses all the pieces of output from gamess: the stdout in
    *gamessout* Scratch files are not yet considered at this moment.
    """
    qcvars, calc_mol, calc_grad, module = harvest_output(gamessout)

    datasections = {}
    if outfiles.get("gamess.dat"):
        datasections = harvest_datfile(outfiles["gamess.dat"])

    calc_hess = None
    if "$HESS" in datasections:
        calc_hess = load_hessian(datasections["$HESS"], dtype="gamess")
        if np.count_nonzero(calc_hess) == 0:
            calc_hess = None

    # Sometimes the hierarchical setting of CURRENT breaks down
    if method == "ccsd+t(ccsd)":
        qcvars["CURRENT CORRELATION ENERGY"] = qcvars["CCSD+T(CCSD) CORRELATION ENERGY"]
        qcvars["CURRENT ENERGY"] = qcvars["CCSD+T(CCSD) TOTAL ENERGY"]

    if calc_mol:
        qcvars["NUCLEAR REPULSION ENERGY"] = str(round(calc_mol.nuclear_repulsion_energy(), 8))
        if in_mol:
            if abs(calc_mol.nuclear_repulsion_energy() - in_mol.nuclear_repulsion_energy()) > 1.0e-3:
                raise ValueError(
                    f"""GAMESS outfile (NRE: {calc_mol.nuclear_repulsion_energy()}) inconsistent with AtomicInput.molecule (NRE: {in_mol.nuclear_repulsion_energy()})."""
                )

        # Frame considerations
        # * `in_mol` built with deliberation and with all fields accessible.
        # * `calc_mol` has the internally consistent geometry frame but otherwise dinky (geom & symbols & maybe chgmult).
        if in_mol.fix_com and in_mol.fix_orientation:
            # Impose input frame if important as signalled by fix_*=T
            return_mol = in_mol
            _, data = calc_mol.align(in_mol, atoms_map=False, mols_align=True, run_mirror=True, verbose=0)
            mill = data["mill"]

        else:
            return_mol, _ = in_mol.align(calc_mol, atoms_map=False, mols_align=True, run_mirror=True, verbose=0)
            mill = qcel.molutil.compute_scramble(
                len(in_mol.symbols), do_resort=False, do_shift=False, do_rotate=False, do_mirror=False
            )  # identity AlignmentMill

        return_grad = None
        if calc_grad is not None:
            return_grad = mill.align_gradient(calc_grad)

        return_hess = None
        if calc_hess is not None:
            return_hess = mill.align_hessian(np.array(calc_hess))

    else:
        raise ValueError("""No coordinate information extracted from gamess output.""")

    return qcvars, return_hess, return_grad, return_mol, module


def harvest_datfile(datfile: str) -> Dict[str, str]:
    sections = datfile.split(r"$END")
    goodies = {}

    for i, sec in enumerate(sections):
        lsec = sec.split("\n")
        for iln, ln in enumerate(lsec):
            if ln.strip():
                key = ln.strip()
                break
        goodies[key] = "\n".join(lsec[iln + 1 :])

    return goodies


def harvest_output(outtext):
    """Function to separate portions of a gamess output file *outtext*,
    divided by "Step".
    """
    pass_qcvar = []
    pass_coord = []
    pass_grad = []

    for outpass in re.split(
        # fmt: off
        r"^\s+" + r"--------" + r"NSERCH:" + r"([1-9][0-9][0-9][0-9]*)" + r"\s*" +
        r"^\s+" + r"--------",
        # fmt: on
        outtext,
        re.MULTILINE,
    ):
        qcvar, gamesscoord, gamessgrad, module = harvest_outfile_pass(outpass)
        pass_qcvar.append(qcvar)
        pass_coord.append(gamesscoord)
        pass_grad.append(gamessgrad)

    retindx = -1 if pass_coord[-1] else -2
    return pass_qcvar[retindx], pass_coord[retindx], pass_grad[retindx], module


def harvest_outfile_pass(outtext):
    """Function to read gamess output file *outtext* and parse important
    quantum chemical information from it in
    """
    qcvar = PreservingDict()
    qcvar_coord = None
    qcvar_grad = None
    module = None

    NUMBER = r"(?x:" + regex.NUMBER + ")"

    # If calculation fail to converge
    mobj = re.search(r"^\s+" + r"(?:GAMESS TERMINATED ABNORMALLY)" + r"\s*$", outtext, re.MULTILINE)
    if mobj:
        logger.debug("GAMESS TERMINATED ABNORMALLY")

    # If calculation converged
    else:
        mobj = re.search(
            r"^\s+" + r"(?:            TOTAL ENERGY)" + r"\s+=\s*" + NUMBER + r"s*$", outtext, re.MULTILINE
        )
        if mobj:
            logger.debug("matched gamess_RHF energy")
            qcvar["HF TOTAL ENERGY"] = mobj.group(1)
            qcvar["SCF TOTAL ENERGY"] = mobj.group(1)

        # Process NRE
        mobj = re.search(
            r"^\s+" + r"(?:   NUCLEAR REPULSION ENERGY)" + r"\s+=\s*" + NUMBER + r"\s*$", outtext, re.MULTILINE
        )
        if mobj:
            logger.debug("matched NRE")
            qcvar["NUCLEAR REPULSION ENERGY"] = mobj.group(1)

        # Process calcinfo
        mobj = re.search(
            # fmt: off
            r"^\s+" + r"NUMBER OF OCCUPIED ORBITALS \(ALPHA\)" + r"\s+=\s+" + r"(?P<nao>\d+)" + r"\s*" +
            r"^\s+" + r"NUMBER OF OCCUPIED ORBITALS \(BETA \)" + r"\s+=\s+" + r"(?P<nbo>\d+)" + r"\s*" +
            r"^\s+" + r"TOTAL NUMBER OF ATOMS" + r"\s+=\s+" + r"(?P<nat>\d+)" + r"\s*",
            # fmt: on
            outtext,
            re.MULTILINE,
        )
        if mobj:
            logger.debug("matched calcinfo")
            qcvar["N ALPHA ELECTRONS"] = mobj.group("nao")
            qcvar["N BETA ELECTRONS"] = mobj.group("nbo")

        mobj = re.search(
            # fmt: off
            r"^\s+" + r"TOTAL NUMBER OF MOS IN VARIATION SPACE=" + r"\s+" + r"(?P<nmo>\d+)" + r"\s*$",
            # fmt: on
            outtext,
            re.MULTILINE,
        )
        if mobj:
            logger.debug("matched calcinfo 2")
            qcvar["N MOLECULAR ORBITALS"] = mobj.group("nmo")
            qcvar["N BASIS FUNCTIONS"] = mobj.group("nmo")  # TODO BAD
        else:
            mobj2 = re.search(
                # fmt: off
                r"^\s+" + r"NUMBER OF CORE -A-  ORBITALS" + r"\s+=\s+" + r"(?P<core_nao>\d+)" + r"\s*" +
                r"^\s+" + r"NUMBER OF CORE -B-  ORBITALS" + r"\s+=\s+" + r"(?P<core_nbo>\d+)" + r"\s*" +
                r"^\s+" + r"NUMBER OF OCC. -A-  ORBITALS" + r"\s+=\s+" + r"(?P<occ_nao>\d+)" + r"\s*" +
                r"^\s+" + r"NUMBER OF OCC. -B-  ORBITALS" + r"\s+=\s+" + r"(?P<occ_nbo>\d+)" + r"\s*" +
                r"^\s+" + r"NUMBER OF MOLECULAR ORBITALS" + r"\s+=\s+" + r"(?P<nmo>\d+)" + r"\s*" +
                r"^\s+" + r"NUMBER OF   BASIS  FUNCTIONS" + r"\s+=\s+" + r"(?P<nbf>\d+)" + r"\s*",
                # fmt: on
                outtext,
                re.MULTILINE,
            )
            if mobj2:
                logger.debug("matched calcinfo 3")
                qcvar["N ALPHA ELECTRONS"] = mobj2.group("occ_nao")
                qcvar["N BETA ELECTRONS"] = mobj2.group("occ_nbo")
                qcvar["N MOLECULAR ORBITALS"] = mobj2.group("nmo")
                qcvar["N BASIS FUNCTIONS"] = mobj2.group("nbf")

        # Process MP2
        mobj = re.search(
            # fmt: off
            r"^\s+" + r"MP2 CONTROL INFORMATION" + r"\s*" +
            r"^\s+" + r"-+" + r"\s*" +
            r"^\s+" + r"NACORE" + r"\s+=\s+" + r"\d+" + r"\s+" + r"NBCORE" + r"\s+=\s+" + r"\d+" + r"\s*" +
            r"^\s+" + r"LMOMP2" + r"\s+=\s+" + r"\w+" + r"\s+" + r"AOINTS" + r"\s+=\s+" + r"\w+" + r"\s*" +
            r"^\s+" + r"METHOD" + r"\s+=\s+" + r"\d+" + r"\s+" + r"NWORD" + r"\s+=\s+" + r"\d+" + r"\s*" +
            r"^\s+" + r"MP2PRP" + r"\s+=\s+" + r"\w+" + r"\s+" + r"OSPT" + r"\s+=\s+" + r"\w+"+ r"\s*" +
            r"^\s+" + r"CUTOFF" + r"\s+=\s+" + NUMBER + r"\s+" + r"CPHFBS" + r"\s+=\s+" + r"\w+"+ r"\s*" +
            r"^\s+" + r"CODE" + r"\s+=\s+" + r"(?P<code>\w+)"  + r"\s*$",
            # fmt: on
            outtext,
            re.MULTILINE,
        )
        if mobj:
            logger.debug("matched mp2 module")
            module = mobj.group("code").lower()

        mobj = re.search(
            # fmt: off
            r'^\s+' + r'RESULTS OF MOLLER-PLESSET 2ND ORDER CORRECTION ARE\n'
            r'^\s+' + r'E\(0\)' + r'=\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'E\(1\)' + r'=\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'E\(2\)' + r'=\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'E\(MP2\)' + r'=\s+' + NUMBER + r'\s*$',
            # fmt: on
            outtext,
            re.MULTILINE,
        )
        if mobj:
            logger.debug("matched mp2 a")
            qcvar["MP2 CORRELATION ENERGY"] = mobj.group(3)
            qcvar["MP2 TOTAL ENERGY"] = mobj.group(4)
            mobj3 = re.search(r"\s+[RU]HF\s*SCF\s*CALCULATION", outtext, re.MULTILINE)
            if mobj3:
                qcvar[f"MP2 DOUBLES ENERGY"] = mobj.group(3)

        mobj = re.search(
            # fmt: off
            r'^\s+' + r'RESULTS OF 2ND-ORDER ZAPT CORRECTION' + r'\s*' +
            r'^\s+' + r'E\(HF\)   ' + r'=\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'E\(ZAPT\) ' + r'=\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'-----------------------------------' + r'\s*' +
            r'^\s+' + r'E\(MP2\)  ' + r'=\s+' + NUMBER + r'\s*$',
            # fmt: on
            outtext,
            re.MULTILINE,
        )
        if mobj:
            logger.debug("matched mp2 b")
            qcvar["ZAPT2 CORRELATION ENERGY"] = mobj.group(2)
            qcvar["ZAPT2 TOTAL ENERGY"] = mobj.group(3)

        mobj = re.search(
            # fmt: off
            r'^\s+' + r'RESULTS OF MOLLER-PLESSET 2ND ORDER CORRECTION ARE\n'
            r'^\s+' + r'E\(0\)=' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'E\(1\)=' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'E\(2\)=' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'E\(MP2\)=' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'SPIN-COMPONENT-SCALED MP2 RESULTS ARE' + r'\s*' +
            r'^\s+' + r'E\(2S\)=' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'E\(2T\)=' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'E\(2ST\)=' + r'\s+' + NUMBER + r'\s*',
            # fmt: on
            outtext,
            re.MULTILINE,
        )
        if mobj:
            logger.debug("matched mp2 rhf c")
            qcvar["HF TOTAL ENERGY"] = mobj.group(1)
            qcvar["MP2 SINGLES ENERGY"] = mobj.group(2)
            qcvar["MP2 DOUBLES ENERGY"] = mobj.group(3)
            qcvar["MP2 TOTAL ENERGY"] = mobj.group(4)
            qcvar["MP2 OPPOSITE-SPIN CORRELATION ENERGY"] = mobj.group(5)
            qcvar["MP2 SAME-SPIN CORRELATION ENERGY"] = mobj.group(6)

        mobj = re.search(
            # fmt: off
            r'^\s+' + r'SINGLE EXCITATION CONTRIBUTION' + r'\s*' +
            r'^\s+' + r'ALPHA' + r'\s*' + NUMBER + r'\s*' +
            r'^\s+' + r'BETA' + r'\s*' + NUMBER + r'\s*' +
            r'^\s+' + r'DOUBLE EXCITATION CONTRIBUTION' + r'\s*' +
            r'^\s+' + NUMBER + r'\s*' +
            r'^\s+' +
            r'^\s+' + r'RESULTS OF MOLLER-PLESSET 2ND ORDER CORRECTION ARE' + r'\s*$',
            # fmt: on
            outtext,
            re.MULTILINE,
        )
        if mobj:
            logger.debug("matched mp2 rohf d")
            qcvar["MP2 SINGLES ENERGY"] = Decimal(mobj.group(1)) + Decimal(mobj.group(2))
            qcvar["MP2 DOUBLES ENERGY"] = mobj.group(3)

        mobj = re.search(r"^\s+" + "UHF-MP2 CALCULATION", outtext, re.MULTILINE)
        if mobj:
            logger.debug("matched mp2 uhf e")
            qcvar["MP2 SINGLES ENERGY"] = "0.0"

        mobj = re.search(
            # fmt: off
            r"^\s+" + r"E\(SCF\)" + r"\s*=\s+" + NUMBER + r"\s*" +
            r"^\s+" + r"ZAPT E\(2\)" + r"\s*=\s+" + NUMBER + r"\s*" +
            r"^\s+" + r"E\(MP2\)" + r"\s*=\s+" + NUMBER + r"\s*",
            # fmt: on
            outtext,
            re.MULTILINE,
        )
        if mobj:
            logger.debug("matched mp2 rohf f")
            qcvar["HF TOTAL ENERGY"] = mobj.group(1)
            qcvar["MP2 CORRELATION ENERGY"] = mobj.group(2)
            qcvar["MP2 TOTAL ENERGY"] = mobj.group(3)

        mobj = re.search(
            # fmt: off
            r'^\s+' + r'RESULTS OF MOLLER-PLESSET 2ND ORDER CORRECTION ARE\n'
            r'^\s+' + r'E\(SCF\)=' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'E\(1\)=' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'E\(2\)=' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'E\(MP2\)=' + r'\s+' + NUMBER + r'\s*',
            # fmt: on
            outtext,
            re.MULTILINE,
        )
        if mobj:
            logger.debug("matched mp2 g llel")
            qcvar["HF TOTAL ENERGY"] = mobj.group(1)
            qcvar["MP2 SINGLES ENERGY"] = mobj.group(2)
            qcvar["MP2 DOUBLES ENERGY"] = mobj.group(3)
            qcvar["MP2 CORRELATION ENERGY"] = mobj.group(3)
            qcvar["MP2 TOTAL ENERGY"] = mobj.group(4)

        mobj = re.search(
            # fmt: off
            r'^\s+' + r'RESULTS OF MOLLER-PLESSET 2ND ORDER CORRECTION ARE\n'
            r'^\s+' + r'E\(SCF\)=' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'E\(2\)=' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'E\(MP2\)=' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'SPIN-COMPONENT-SCALED MP2 RESULTS ARE' + r'\s*' +
            r'^\s+' + r'E\(2S\)=' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'E\(2T\)=' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'E\(2ST\)=' + r'\s+' + NUMBER + r'\s*',
            # fmt: on
            outtext,
            re.MULTILINE,
        )
        if mobj:
            logger.debug("matched mp2 rhf h llel")
            qcvar["HF TOTAL ENERGY"] = mobj.group(1)
            qcvar["MP2 SINGLES ENERGY"] = "0.0"
            qcvar["MP2 DOUBLES ENERGY"] = mobj.group(2)
            qcvar["MP2 CORRELATION ENERGY"] = mobj.group(2)
            qcvar["MP2 TOTAL ENERGY"] = mobj.group(3)
            qcvar["MP2 OPPOSITE-SPIN CORRELATION ENERGY"] = mobj.group(4)
            qcvar["MP2 SAME-SPIN CORRELATION ENERGY"] = mobj.group(5)

        # Process CCSD
        mobj = re.search(
            # fmt: off
            r'^\s+' + r'SUMMARY OF RESULTS' + r'\s+' + r'\n' +
            r'^\s+' + r'REFERENCE ENERGY:' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'MBPT\(2\) ENERGY:' + r'\s+' + NUMBER + r'\s*' + r'CORR.E=\s+' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r"(?P<mtd>CCSD|LCCD|CCD)" + r"\s+" r'ENERGY:' + r'\s+' + NUMBER + r'\s*' + r'CORR.E=\s+' + r'\s+' + NUMBER + r'\s*$',
            # fmt: on
            outtext,
            re.MULTILINE,
        )
        if mobj:
            logger.debug("matched rhf ccsd/lccd/ccd")
            mtd = mobj.group("mtd")
            qcvar["HF TOTAL ENERGY"] = mobj.group(1)
            qcvar["SCF TOTAL ENERGY"] = mobj.group(1)
            qcvar["MP2 CORRELATION ENERGY"] = mobj.group(3)
            qcvar["MP2 DOUBLES ENERGY"] = mobj.group(3)
            qcvar["MP2 TOTAL ENERGY"] = mobj.group(2)
            qcvar[f"{mtd} DOUBLES ENERGY"] = mobj.group(6)
            qcvar[f"{mtd} CORRELATION ENERGY"] = mobj.group(6)
            qcvar[f"{mtd} TOTAL ENERGY"] = mobj.group(5)

        mobj = re.search(
            # fmt: off
            r'^\s+' + r'SUMMARY OF CCSD RESULTS' + r'\s+' + r'\n' +
            r'^\s+' + r'REFERENCE ENERGY:' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'CCSD ENERGY:'   + r'\s+' + NUMBER + r'\s*' + r'CORR. E=\s+' + r'\s+' + NUMBER + r'\s*$',
            # fmt: on
            outtext,
            re.MULTILINE,
        )
        if mobj:
            logger.debug("matched rohf ccsd")
            qcvar["SCF TOTAL ENERGY"] = mobj.group(1)
            qcvar["CCSD CORRELATION ENERGY"] = mobj.group(3)
            qcvar["CCSD TOTAL ENERGY"] = mobj.group(2)

        # Process CR-CC(2,3)
        mobj = re.search(
            # fmt: off
            r'^\s+' + r'CCSD                       ENERGY:'       + r'\s+' + NUMBER + r'\s*' + r'CORR.E=\s+' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'CR-CC\(2,3\),A OR CCSD\(2\)_T  ENERGY:'   + r'\s+' + NUMBER + r'\s*' + r'CORR.E=\s+' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'CR-CC\(2,3\) OR CR-CCSD\(T\)_L ENERGY:'   + r'\s+' + NUMBER + r'\s*' + r'CORR.E=\s+' + r'\s+' + NUMBER + r'\s*',
            # fmt: on
            outtext,
            re.MULTILINE,
        )
        if mobj:
            logger.debug("matched cc-cr(2,3)")
            qcvar["CCSD TOTAL ENERGY"] = mobj.group(1)
            qcvar["CCSD CORRELATION ENERGY"] = mobj.group(2)
            qcvar["CR-CC(2,3),A TOTAL ENERGY"] = mobj.group(3)
            qcvar["CR-CC(2,3),A CORRELATION ENERGY"] = mobj.group(4)
            qcvar["CR-CC(2,3) TOTAL ENERGY"] = mobj.group(5)
            qcvar["CR-CC(2,3) CORRELATION ENERGY"] = mobj.group(6)

        # Process CCSD(T)
        mobj = re.search(
            # fmt: off
            r'^\s+' + r'SUMMARY OF RESULTS' + r'\s+' + r'\n' +
            r'^\s+' + r'REFERENCE ENERGY:' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'MBPT\(2\) ENERGY:' + r'\s+' + NUMBER + r'\s*' + r'CORR.E=\s+' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'CCSD    ENERGY:'   + r'\s+' + NUMBER + r'\s*' + r'CORR.E=\s+' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'CCSD\[T\] ENERGY:' + r'\s+' + NUMBER + r'\s*' + r'CORR.E=\s+' + r'\s+' + NUMBER + r'\s*' +
            r'^\s+' + r'CCSD\(T\) ENERGY:' + r'\s+' + NUMBER + r'\s*' + r'CORR.E=\s+' + r'\s+' + NUMBER + r'\s*$',
            # fmt: on
            outtext,
            re.MULTILINE,
        )
        if mobj:
            logger.debug("matched ccsd(t)")
            qcvar["HF TOTAL ENERGY"] = mobj.group(1)
            qcvar["SCF TOTAL ENERGY"] = mobj.group(1)
            qcvar["CCSD CORRELATION ENERGY"] = mobj.group(5)
            qcvar["CCSD TOTAL ENERGY"] = mobj.group(4)
            qcvar["(T) CORRECTION ENERGY"] = Decimal(mobj.group(8)) - Decimal(mobj.group(4))
            # qcvar['CCSD(T) CORRELATION ENERGY'] = Decimal(mobj.group(5)) - qcvar['SCF TOTAL ENERGY']
            qcvar["CCSD+T(CCSD) CORRELATION ENERGY"] = mobj.group(7)
            qcvar["CCSD+T(CCSD) TOTAL ENERGY"] = mobj.group(6)
            qcvar["CCSD(T) CORRELATION ENERGY"] = mobj.group(9)
            qcvar["CCSD(T) TOTAL ENERGY"] = mobj.group(8)

        # Process CISD
        mobj = re.search(
            # fmt: off
            r"^\s+" + r"PROPERTY VALUES FOR THE " + r"(RHF|ROHF)" + r"\s+" + r"SELF-CONSISTENT FIELD WAVEFUNCTION" + r"\s*" +
            r"(?:.*?)" +
            r"^\s+" + r"ENERGY COMPONENTS" + r"\s*" +
            r"(?:.*?)" +
            r"^\s*" + r"(TOTAL ENERGY =)\s+" + r"(?P<hf>" + NUMBER + r")" + r"\s*" +
            r"(?:.*?)" +
            r"^\s+" + r"(?P<module>(FSOCI|GUGA))" + r"\s+" + r"CI PROPERTIES...FOR THE WAVEFUNCTION OF STATE    1" + r"\s*" +
            r"(?:.*?)" +
            r"^\s+" + r"ENERGY COMPONENTS" + r"\s*" +
            r"(?:.*?)" +
            r"^\s*" + r"(TOTAL ENERGY =)\s+" + r"(?P<ci>" + NUMBER + r")" + r"\s*$",
            # fmt: on
            outtext,
            re.MULTILINE | re.DOTALL,
        )
        if mobj:
            logger.debug("matched cisd fsoci/guga", mobj.groupdict())
            module = mobj.group("module").lower()
            qcvar["CISD CORRELATION ENERGY"] = Decimal(mobj.group("ci")) - Decimal(mobj.group("hf"))
            qcvar["CISD TOTAL ENERGY"] = mobj.group("ci")
            qcvar["CI TOTAL ENERGY"] = mobj.group("ci")

        # Process FCI
        mobj = re.search(
            # fmt: off
            r'^\s+' + r'ALDET CI PROPERTIES...FOR THE WAVEFUNCTION OF STATE' + r'\s+' + r'(?P<state>\d+)' + r'\s*' +
            r'^\s+' + r'USING THE EXPECTATION VALUE DENSITY' + r'\s*' +
            r'(?:.*?)' +
            r'^\s+' + r'TOTAL ENERGY =' + r'\s+' + NUMBER + r'\s*$',
            # fmt: on
            outtext,
            re.MULTILINE | re.DOTALL,
        )
        if mobj:
            logger.debug("matched fci")
            qcvar["FCI TOTAL ENERGY"] = mobj.group(2)
            qcvar["CI TOTAL ENERGY"] = mobj.group(2)

        # Process EOM-CCSD
        mobj = re.search(
            # fmt: off
            r'^\s+' + r'---- SUMMARY OF EOM-CCSD CALCULATIONS ----' +
            r'\s+' + r'\n' +
            r'^\s+' + r'EXCITATION      EXCITATION      TOTAL' + r'\s*' +
            r'^\s+' + r'SYMMETRY     ENERGY \(H\)      ENERGY \(EV\)     ENERGY \(H\)          ITERATIONS' + r'\s*' +
            r'^\s+' + r'A' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s+' + r'CONVERGED\s+' +
            r'^\s+' + r'A' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s+' + r'CONVERGED\s+' +
            r'^\s+' + r'A' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s+' + r'CONVERGED\s+' +
            r'^\s+' + r'A' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s+' + r'CONVERGED\s+',
            # fmt: on
            outtext,
            re.MULTILINE,
        )
        if mobj:
            logger.debug("matched eom-ccsd")
            qcvar["EOM-CCSD ROOT 1 EXCITATION ENERGY"] = mobj.group(1)
            qcvar["EOM-CCSD ROOT 2 EXCITATION ENERGY"] = mobj.group(4)
            qcvar["EOM-CCSD ROOT 3 EXCITATION ENERGY"] = mobj.group(7)
            qcvar["EOM-CCSD ROOT 4 EXCITATION ENERGY"] = mobj.group(10)
            qcvar["EOM-CCSD ROOT 1 TOTAL ENERGY"] = mobj.group(3)
            qcvar["EOM-CCSD ROOT 2 TOTAL ENERGY"] = mobj.group(6)
            qcvar["EOM-CCSD ROOT 3 TOTAL ENERGY"] = mobj.group(9)
            qcvar["EOM-CCSD ROOT 4 TOTAL ENERGY"] = mobj.group(12)

        # Process DFT
        # mobj = re.search(
        #     r'^\s+' + r'DFT EXCHANGE + CORRELATION ENERGY' + r'=\s+' + NUMBER + r'\s*$'
        #     ,outtext, re.MULTILINE)
        mobj = re.search(
            r"^\s+" + r"DFT EXCHANGE \+ CORRELATION ENERGY" + r"\s+=\s*" + NUMBER + r"\s*$", outtext, re.MULTILINE
        )
        if mobj:
            logger.debug("matched dft xc")
            qcvar["DFT XC ENERGY"] = mobj.group(1)

        # Process Geometry
        mobj = re.search(
            # fmt: off
            r'^\s+' + r'ATOM      ATOMIC                      COORDINATES \(BOHR\)' + r'\s*' +
            r'^\s+' + r'CHARGE         X                   Y                   Z'+ r'\s*' +
            r'((?:\s+([A-Za-z]\w*)\s+\d+\.\d+\s+[-+]?\d+\.\d+\s+[-+]?\d+\.\d+\s+[-+]?\d+\.\d+\s*\n)+)' +
            r"\s*$",
            # fmt: on
            outtext,
            re.MULTILINE | re.IGNORECASE,
        )
        if mobj:
            logger.debug("matched geom")
            molxyz = "%d bohr\n\n" % len(mobj.group(1).splitlines())
            for line in mobj.group(1).splitlines():
                lline = line.split()
                chg_on_center = int(float(lline[-4]))
                if chg_on_center > 0:
                    tag = f"{lline[-5]}"
                else:
                    tag = f"@{lline[-5].strip()}"
                molxyz += "%s %16s %16s %16s\n" % (tag, lline[-3], lline[-2], lline[-1])
            qcvar_coord = Molecule(
                validate=False,
                **qcel.molparse.to_schema(
                    qcel.molparse.from_string(molxyz, dtype="xyz+", fix_com=True, fix_orientation=True)["qm"], dtype=2
                ),
            )

        # Process Gradient
        mobj = re.search(
            # fmt: off
            r'^\s+' + r'GRADIENT OF THE ENERGY' + r'\s*'+
            r'^\s+' + r'----------------------' + r'\s*'+
            r'\s+' + r'\n'+
            r'^\s+' + r'UNITS ARE HARTREE/BOHR    E\'X               E\'Y               E\'Z' + r'\s*' +
            r'((?:\s+([1-9][0-9]*)+\s+([A-Za-z]\w*)+\s+[-+]?\d+\.\d+\s+[-+]?\d+\.\d+\s+[-+]?\d+\.\d+\s*\n)+)' +
            r'\s*$',
            # fmt: off
                outtext, re.MULTILINE
        )
        if mobj:
            logger.debug("matched gradient - after")
            atoms = []
            qcvar_grad = []
            for line in mobj.group(1).splitlines():
                lline = line.split()
                if lline == []:
                    pass
                else:
                    logger.debug("printing gradient")
                    atoms.append(lline[1])
                    qcvar_grad.append([float(lline[-3]), float(lline[-2]), float(lline[-1])])
            qcvar_grad = np.array(qcvar_grad).reshape((-1, 3))

        # Process SCF blocks
        mobj = re.search(
            # fmt: off
            r'^\s+' + r'PROPERTIES FOR THE .* DFT FUNCTIONAL .* DENSITY MATRIX' +
            r'(.*)' +
            r'\.\.\.\.\.\. END OF PROPERTY EVALUATION \.\.\.\.\.\.',
            # fmt: on
            outtext,
            re.MULTILINE | re.DOTALL,
        )
        if mobj:
            logger.debug("matched dft props")
            prop_block = mobj.group(1)
            mobj_energy = re.search(
                # fmt: off
                r'\s+ENERGY COMPONENTS\s*\n' +
                r'\s+-----------------\s*\n' +
                r'\s*\n' +
                r'\s+WAVEFUNCTION NORMALIZATION =.*\n' +
                r'\s*\n' +
                r'\s+ONE ELECTRON ENERGY =\s+' + NUMBER + r'\n' +
                r'\s+TWO ELECTRON ENERGY =\s+' + NUMBER + r'\n' +
                r'\s+NUCLEAR REPULSION ENERGY =\s+' + NUMBER + r'\n' +
                r'\s+------------------\s*\n' +
                r'\s+TOTAL ENERGY =\s+' + NUMBER+r'\n',
                # fmt: on
                prop_block
            )
            if mobj_energy:
                qcvar["ONE-ELECTRON ENERGY"] = mobj_energy.group(1)
                qcvar["TWO-ELECTRON ENERGY"] = mobj_energy.group(2)
                qcvar["SCF TOTAL ENERGY"] = mobj_energy.group(4)
                qcvar["DFT TOTAL ENERGY"] = mobj_energy.group(4)

            mobj_dipole = re.search(
                # fmt: off
                r'\s+ELECTROSTATIC MOMENTS\s*\n'
                r'\s+---------------------\s*\n'
                r'\s*\n'
                r'\s*POINT\s+1\s+X\s+Y\s+Z\s+\(BOHR\)\s+CHARGE\s*\n'
                r'.*\n'
                r'\s*DX\s+DY\s+DZ\s+/D/\s+\(DEBYE\)\s*\n'
                r'\s*' + NUMBER + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s*\n',
                # fmt: on
                prop_block
            )
            if mobj_dipole:
                d2au = Decimal(qcel.constants.conversion_factor("debye", "e * bohr"))
                qcvar["SCF DIPOLE"] = d2au * np.array(
                    [Decimal(mobj_dipole.group(1)), Decimal(mobj_dipole.group(2)), Decimal(mobj_dipole.group(3))]
                )

    # Process CURRENT Energies
    if "HF TOTAL ENERGY" in qcvar:
        qcvar["CURRENT REFERENCE ENERGY"] = qcvar["HF TOTAL ENERGY"]
        qcvar["CURRENT ENERGY"] = qcvar["HF TOTAL ENERGY"]

    if "MP2 TOTAL ENERGY" in qcvar and "MP2 CORRELATION ENERGY" in qcvar:
        qcvar["CURRENT CORRELATION ENERGY"] = qcvar["MP2 CORRELATION ENERGY"]
        qcvar["CURRENT ENERGY"] = qcvar["MP2 TOTAL ENERGY"]

    if "ZAPT2 TOTAL ENERGY" in qcvar and "ZAPT2 CORRELATION ENERGY" in qcvar:
        qcvar["CURRENT CORRELATION ENERGY"] = qcvar["ZAPT2 CORRELATION ENERGY"]
        qcvar["CURRENT ENERGY"] = qcvar["ZAPT2 TOTAL ENERGY"]

    if "CISD TOTAL ENERGY" in qcvar and "CISD CORRELATION ENERGY" in qcvar:
        qcvar["CURRENT CORRELATION ENERGY"] = qcvar["CISD CORRELATION ENERGY"]
        qcvar["CURRENT ENERGY"] = qcvar["CISD TOTAL ENERGY"]

    if "LCCD TOTAL ENERGY" in qcvar and "LCCD CORRELATION ENERGY" in qcvar:
        qcvar["CURRENT CORRELATION ENERGY"] = qcvar["LCCD CORRELATION ENERGY"]
        qcvar["CURRENT ENERGY"] = qcvar["LCCD TOTAL ENERGY"]

    if "CCD TOTAL ENERGY" in qcvar and "CCD CORRELATION ENERGY" in qcvar:
        qcvar["CURRENT CORRELATION ENERGY"] = qcvar["CCD CORRELATION ENERGY"]
        qcvar["CURRENT ENERGY"] = qcvar["CCD TOTAL ENERGY"]

    if "CCSD TOTAL ENERGY" in qcvar and "CCSD CORRELATION ENERGY" in qcvar:
        qcvar["CURRENT CORRELATION ENERGY"] = qcvar["CCSD CORRELATION ENERGY"]
        qcvar["CURRENT ENERGY"] = qcvar["CCSD TOTAL ENERGY"]

    if "CR-CC(2,3) TOTAL ENERGY" in qcvar and "CR-CC(2,3) CORRELATION ENERGY" in qcvar:
        qcvar["CURRENT CORRELATION ENERGY"] = qcvar["CR-CC(2,3) CORRELATION ENERGY"]
        qcvar["CURRENT ENERGY"] = qcvar["CR-CC(2,3) TOTAL ENERGY"]

    if "CCSD+T(CCSD) TOTAL ENERGY" in qcvar and "CCSD+T(CCSD) CORRELATION ENERGY" in qcvar:
        qcvar["CURRENT CORRELATION ENERGY"] = qcvar["CCSD+T(CCSD) CORRELATION ENERGY"]
        qcvar["CURRENT ENERGY"] = qcvar["CCSD+T(CCSD) TOTAL ENERGY"]

    if "CCSD(T) TOTAL ENERGY" in qcvar and "CCSD(T) CORRELATION ENERGY" in qcvar:
        qcvar["CURRENT CORRELATION ENERGY"] = qcvar["CCSD(T) CORRELATION ENERGY"]
        qcvar["CURRENT ENERGY"] = qcvar["CCSD(T) TOTAL ENERGY"]

    if "DFT TOTAL ENERGY" in qcvar:
        qcvar["CURRENT REFERENCE ENERGY"] = qcvar["DFT TOTAL ENERGY"]
        qcvar["CURRENT ENERGY"] = qcvar["DFT TOTAL ENERGY"]
        qcvar.pop("HF TOTAL ENERGY")

    if "FCI TOTAL ENERGY" in qcvar:  # and 'FCI CORRELATION ENERGY' in qcvar:
        qcvar["CURRENT ENERGY"] = qcvar["FCI TOTAL ENERGY"]

    qcvar[f"N ATOMS"] = len(qcvar_coord.symbols)

    return qcvar, qcvar_coord, qcvar_grad, module
